from popfinder.dataloader import GeneticData
from popfinder.classifier import PopClassifier
from popfinder.regressor import PopRegressor
from popfinder._neural_networks import ClassifierNet
from popfinder._neural_networks import RegressorNet
import pytest
import numpy as np
import pandas as pd
import torch
import os
import shutil
import re
from sklearn.preprocessing import LabelEncoder

TEST_OUTPUT_FOLDER = "tests/test_outputs"

# Empty test output folder if it exists
if os.path.exists(TEST_OUTPUT_FOLDER):
    shutil.rmtree(TEST_OUTPUT_FOLDER)

# Recreate test output folder
os.mkdir(TEST_OUTPUT_FOLDER)

# Test dataloader class
def test_genetic_data_inputs():

    gen_dat = GeneticData()
    assert isinstance(gen_dat, GeneticData)
    with pytest.raises(ValueError, match="genetic_data is None"):
        gen_dat.read_data()

    gen_dat = GeneticData(genetic_data="tests/test_data/test.vcf")  
    with pytest.raises(ValueError, match="sample_data is None"):
        gen_dat.read_data()

    with pytest.raises(ValueError, match="Path to genetic_data does not exist"):
        GeneticData(genetic_data="bad/path.vcf",
                            sample_data="tests/test_data/testNA.txt")

    with pytest.raises(ValueError, match="genetic_data must have extension 'zarr', 'vcf', or 'hdf5'"):
        GeneticData(genetic_data="tests/test_data/testNA.txt",
                            sample_data="tests/test_data/testNA.txt")

    with pytest.raises(ValueError, match="Path to sample_data does not exist"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data="bad/path.txt")

    with pytest.raises(ValueError, match="sample_data must have extension 'txt' or 'tsv'"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data="tests/test_data/test.vcf")

    with pytest.raises(ValueError, match="sample_data file does not have correct columns"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data="tests/test_data/test_bad.txt")

    with pytest.raises(ValueError, match="genetic_data must be a string"):
        GeneticData(genetic_data=123,
                            sample_data="tests/test_data/testNA.txt")

    with pytest.raises(ValueError, match="sample_data must be a string"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data=123)

    with pytest.raises(ValueError, match="test_size must be a float"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data="tests/test_data/testNA.txt",
                            test_size="0.2")

    with pytest.raises(ValueError, match="test_size must be between 0 and 1"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data="tests/test_data/testNA.txt",
                            test_size=2.5)

    with pytest.raises(ValueError, match="seed must be an integer"):
        GeneticData(genetic_data="tests/test_data/test.vcf",
                            sample_data="tests/test_data/testNA.txt",
                            test_size=0.2,
                            seed=0.5)             

def test_genetic_data():

    gen_dat = GeneticData(genetic_data="tests/test_data/test.vcf",
                          sample_data="tests/test_data/testNA.txt")
    assert isinstance(gen_dat, GeneticData)

    dat = gen_dat.read_data()
    assert dat.equals(gen_dat.data)

    assert type(gen_dat.label_enc) == LabelEncoder

    assert gen_dat.data.empty == False
    assert gen_dat.knowns.empty == False
    assert gen_dat.unknowns.empty == False
    assert gen_dat.train.empty == False
    assert gen_dat.test.empty == False

def test_update_unknowns():
    gen_dat = GeneticData(genetic_data="tests/test_data/test.vcf",
                        sample_data="tests/test_data/testNA.txt")

    old_data = gen_dat.data.copy()
    old_knowns = gen_dat.knowns.copy()
    old_unknowns = gen_dat.unknowns.copy()

    gen_dat.update_unknowns(new_genetic_data="tests/test_data/test_new_unknowns.vcf",
                            new_sample_data="tests/test_data/test_new_unknowns.txt")

    assert gen_dat.data.equals(old_data) == False
    assert gen_dat.knowns.equals(old_knowns)
    assert gen_dat.unknowns.equals(old_unknowns) == False

# Test classifier class
def test_classifier_inputs():
    
    data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
                        sample_data="tests/test_data/testNA.txt")

    with pytest.raises(TypeError, match="data must be an instance of GeneticData"):
        PopClassifier(data=None)

    with pytest.raises(TypeError, match="random_state must be an integer"):
        PopClassifier(data_obj, random_state=0.5)

    with pytest.raises(TypeError, match="output_folder must be a string"):
        PopClassifier(data_obj, output_folder=123)

def test_classifier_train():

    data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
                    sample_data="tests/test_data/testNA.txt")
    classifier = PopClassifier(data_obj, output_folder=TEST_OUTPUT_FOLDER)
    
    assert isinstance(classifier, PopClassifier)
    assert classifier.data.data.equals(data_obj.data)
    assert classifier.data.knowns.equals(data_obj.knowns)
    assert classifier.data.unknowns.equals(data_obj.unknowns)
    assert classifier.data.train.equals(data_obj.train)
    assert classifier.data.test.equals(data_obj.test)

    with pytest.raises(TypeError, match="epochs must be an integer"):
        classifier.train(epochs=0.5)

    with pytest.raises(TypeError, match="valid_size must be a float"):
        classifier.train(valid_size="0.2")

    with pytest.raises(ValueError, match="valid_size must be between 0 and 1"):
        classifier.train(valid_size=2.5)   

    with pytest.raises(TypeError, match="cv_splits must be an integer"):
        classifier.train(cv_splits="0.2")

    with pytest.raises(TypeError, match="nreps must be an integer"):
        classifier.train(nreps="0.2")

    with pytest.raises(TypeError, match="learning_rate must be a float"):
        classifier.train(learning_rate="0.2")

    with pytest.raises(ValueError, match="learning_rate must be between 0 and 1"):
        classifier.train(learning_rate=2.7)     

    with pytest.raises(TypeError, match="batch_size must be an integer"):
        classifier.train(batch_size=0.5)

    with pytest.raises(TypeError, match="dropout_prop must be a float"):
        classifier.train(dropout_prop="0.2")   

    with pytest.raises(ValueError, match="dropout_prop must be between 0 and 1"):
        classifier.train(dropout_prop=2) 

    classifier.train()
    assert isinstance(classifier.train_history, pd.DataFrame)
    assert classifier.train_history.empty == False
    assert isinstance(classifier.best_model, torch.nn.Module)

    # Output folder should contain one result folder (rep1_boot1) + best model
    f = "rep1_boot1"
    assert "best_model.pt" in os.listdir(classifier.output_folder)
    assert f in os.listdir(classifier.output_folder)
    assert "best_model_split1.pt" in os.listdir(os.path.join(classifier.output_folder, f))
    assert "loss.csv" in os.listdir(os.path.join(classifier.output_folder, f))

    # Reset classifier object
    classifier = PopClassifier(data_obj, output_folder=TEST_OUTPUT_FOLDER)

    # Testing parameter combos
    params = {"jobs": [1, 2],
              "reps": [1, 2],
              "bootstraps": [1,2],
              "splits": [1,2]}
    
    for job in params["jobs"]:
        for rep in params["reps"]:
            for boot in params["bootstraps"]:
                for split in params["splits"]:
                    print(f"Testing job={job}, rep={rep}, boot={boot}, split={split}")
                    classifier.train(cv_splits=split, nreps=rep, bootstraps=boot, jobs=job, epochs=10)
                    f = f"rep{rep}_boot{boot}"
                    assert "best_model.pt" in os.listdir(classifier.output_folder)
                    assert f in os.listdir(classifier.output_folder)
                    assert f"best_model_split{split}.pt" in os.listdir(os.path.join(classifier.output_folder, f))
                    assert "loss.csv" in os.listdir(os.path.join(classifier.output_folder, f))
                    assert len(classifier.train_history["rep"].unique()) == rep
                    assert len(classifier.train_history["bootstrap"].unique()) == boot

    # Non-multiprocessing append to results
    old_train_history = len(classifier.train_history)
    classifier.train(nreps=2, bootstraps=2, epochs = 10, overwrite_results=False)
    assert len(classifier.train_history) == old_train_history + (2 * 2 * 10)

    # Multiprocessing append to results
    old_train_history = len(classifier.train_history)
    classifier.train(jobs=2, nreps=2, bootstraps=2, epochs = 10, overwrite_results=False)
    assert len(classifier.train_history) == old_train_history + (2 * 2 * 10)

def test_classifier_test():

    data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
                    sample_data="tests/test_data/testNA.txt")
    classifier = PopClassifier(data_obj, output_folder=TEST_OUTPUT_FOLDER)
    classifier.train()
    classifier.test()

    # Check outputs
    assert isinstance(classifier.test_results, pd.DataFrame)
    assert classifier.test_results.empty == False
    assert isinstance(classifier.confusion_matrix, np.ndarray)
    assert classifier.confusion_matrix.shape == (5,5)
    assert np.round(classifier.confusion_matrix.sum(), 1) == np.round(5.0, 1)
    assert isinstance(classifier.accuracy, float)
    assert classifier.accuracy > 0.0
    assert classifier.accuracy < 1.0
    assert isinstance(classifier.precision, float)
    assert classifier.precision > 0.0
    assert classifier.precision < 1.0
    assert isinstance(classifier.recall, float)
    assert classifier.recall > 0.0
    assert classifier.recall < 1.0
    assert isinstance(classifier.f1, float)
    assert classifier.f1 > 0.0
    assert classifier.f1 < 1.0

    # Test non-multiprocessing results
    classifier.train(cv_splits=2, nreps=2, bootstraps=2)
    classifier.test(use_best_model=False)
    assert np.sum([x in classifier.test_results.columns for x in ["rep", "split"]]) == 2
    assert len(classifier.test_results["rep"].unique()) == 2
    assert len(classifier.test_results["bootstrap"].unique()) == 2
    assert len(classifier.test_results["split"].unique()) == 2 
    classifier.test()
    assert np.sum([x not in classifier.test_results.columns for x in ["rep", "split"]]) == 2

    # Test multiprocessing results
    classifier.train(cv_splits=2, nreps=2, bootstraps=2, jobs=2)
    classifier.test(use_best_model=False)
    assert np.sum([x in classifier.test_results.columns for x in ["rep", "split"]]) == 2
    assert len(classifier.test_results["rep"].unique()) == 2
    assert len(classifier.test_results["bootstrap"].unique()) == 2
    assert len(classifier.test_results["split"].unique()) == 2 
    classifier.test()
    assert np.sum([x not in classifier.test_results.columns for x in ["rep", "split"]]) == 2

    # Test save
    save_path = os.path.join(classifier.output_folder, "classifier_test_results.csv")
    assert os.path.exists(save_path)
    os.remove(save_path)
    classifier.test(save=False)
    assert not os.path.exists(save_path)

def test_classifier_assign_unknown_and_get_results():

    data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
                    sample_data="tests/test_data/testNA.txt")
    classifier = PopClassifier(data_obj, output_folder=TEST_OUTPUT_FOLDER)
    classifier.train()
    classifier.test()
    unknown_data = classifier.assign_unknown()

    assert unknown_data.equals(classifier.classification)
    assert os.path.exists(os.path.join(classifier.output_folder,
                          "classifier_assignment_results.csv"))
    os.remove(os.path.join(classifier.output_folder,
                            "classifier_assignment_results.csv"))

    class_sum = classifier.get_classification_summary(save=False)
    assert isinstance(class_sum, pd.DataFrame)
    assert not os.path.exists(os.path.join(classifier.output_folder,
                              "classifier_classification_summary.csv"))

    classifier.get_classification_summary(save=True)
    assert os.path.exists(os.path.join(classifier.output_folder,
                          "classifier_classification_summary.csv"))
    os.remove(os.path.join(classifier.output_folder,
                            "classifier_classification_summary.csv"))

    site_rank = classifier.rank_site_importance()
    assert isinstance(site_rank, pd.DataFrame)
    assert site_rank.empty == False
    assert len(classifier.data.data.alleles[0]) == len(site_rank)

def test_classifier_update_unknowns():

    data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
                    sample_data="tests/test_data/testNA.txt")
    classifier = PopClassifier(data_obj, output_folder=TEST_OUTPUT_FOLDER)
    classifier.train()
    classifier.test()
    classifier.assign_unknown()

    old_data = classifier.data.data.copy()
    old_knowns = classifier.data.knowns.copy()
    old_unknowns = classifier.data.unknowns.copy()
    old_classification_samples = classifier.classification.sampleID.copy()

    classifier.update_unknown_samples(new_genetic_data="tests/test_data/test_new_unknowns.vcf",
                                      new_sample_data="tests/test_data/test_new_unknowns.txt")

    assert classifier.data.data.equals(old_data) == False
    assert classifier.data.knowns.equals(old_knowns)
    assert classifier.data.unknowns.equals(old_unknowns) == False

    classifier.assign_unknown()

    assert classifier.classification.sampleID.equals(old_classification_samples) == False

def test_classifier_save_and_load():

    data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
                    sample_data="tests/test_data/testNA.txt")
    classifier = PopClassifier(data_obj, output_folder=TEST_OUTPUT_FOLDER)
    classifier.train()
    classifier.test()
    classifier.assign_unknown()
    classifier.save()

    assert os.path.exists(os.path.join(classifier.output_folder,
                          "classifier.pkl"))

    classifier2 = PopClassifier.load(load_path=os.path.join(classifier.output_folder,
                                "classifier.pkl"))

    assert classifier2.train_history.equals(classifier.train_history)
    assert classifier2.test_results.equals(classifier.test_results)
    assert np.array_equal(classifier2.confusion_matrix, classifier.confusion_matrix)
    assert classifier2.accuracy == classifier.accuracy
    assert classifier2.precision == classifier.precision
    assert classifier2.recall == classifier.recall
    assert classifier2.f1 == classifier.f1
    assert classifier2.classification.equals(classifier.classification)
    assert isinstance(classifier2.best_model, ClassifierNet)
    assert isinstance(classifier.best_model, ClassifierNet)

    os.remove(os.path.join(classifier.output_folder,
                            "classifier.pkl"))


# Test regressor class
# def test_regressor_inputs():

#     data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
#                         sample_data="tests/test_data/testNA.txt")

#     with pytest.raises(TypeError, match="data must be an instance of GeneticData"):
#         PopRegressor(data=None)

#     with pytest.raises(TypeError, match="nboots must be an integer"):
#         PopRegressor(data=data_obj, nboots=0.5)

#     with pytest.raises(TypeError, match="random_state must be an integer"):
#         PopRegressor(data_obj, random_state=0.5)

#     with pytest.raises(TypeError, match="output_folder must be a string"):
#         PopRegressor(data_obj, output_folder=123)

#     with pytest.raises(ValueError, match="output_folder must be a valid directory"):
#         PopRegressor(data_obj, output_folder="bad/path")

# def test_regressor_train():

#     data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
#                     sample_data="tests/test_data/testNA.txt")
#     regressor = PopRegressor(data_obj, output_folder=TEST_OUTPUT_FOLDER)
    
#     assert isinstance(regressor, PopRegressor)
#     assert regressor.data.data.equals(data_obj.data)
#     assert regressor.data.knowns.equals(data_obj.knowns)
#     assert regressor.data.unknowns.equals(data_obj.unknowns)
#     assert regressor.data.train.equals(data_obj.train)
#     assert regressor.data.test.equals(data_obj.test)

#     with pytest.raises(TypeError, match="epochs must be an integer"):
#         regressor.train(epochs=0.5)

#     with pytest.raises(TypeError, match="valid_size must be a float"):
#         regressor.train(valid_size="0.2")

#     with pytest.raises(ValueError, match="valid_size must be between 0 and 1"):
#         regressor.train(valid_size=2.5)   

#     with pytest.raises(TypeError, match="cv_splits must be an integer"):
#         regressor.train(cv_splits="0.2")

#     with pytest.raises(TypeError, match="cv_reps must be an integer"):
#         regressor.train(cv_reps="0.2")

#     with pytest.raises(TypeError, match="learning_rate must be a float"):
#         regressor.train(learning_rate="0.2")

#     with pytest.raises(ValueError, match="learning_rate must be between 0 and 1"):
#         regressor.train(learning_rate=2.7)     

#     with pytest.raises(TypeError, match="batch_size must be an integer"):
#         regressor.train(batch_size=0.5)

#     with pytest.raises(TypeError, match="dropout_prop must be a float"):
#         regressor.train(dropout_prop="0.2")   

#     with pytest.raises(ValueError, match="dropout_prop must be between 0 and 1"):
#         regressor.train(dropout_prop=2) 

#     with pytest.raises(TypeError, match="boot_data must be an instance of GeneticData"):
#         regressor.train(boot_data="0.2")

#     regressor.train()
#     assert isinstance(regressor.train_history, pd.DataFrame)
#     assert regressor.train_history.empty == False
#     assert isinstance(regressor.best_model, torch.nn.Module)

# def test_regressor_test():

#     data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
#                     sample_data="tests/test_data/testNA.txt")
#     regressor = PopRegressor(data_obj, output_folder=TEST_OUTPUT_FOLDER)
#     regressor.train()
#     regressor.test()

#     assert isinstance(regressor.test_results, pd.DataFrame)
#     assert regressor.test_results.empty == False
#     assert isinstance(regressor.median_distance, float)
#     assert isinstance(regressor.mean_distance, float)
#     assert isinstance(regressor.r2_long, float)
#     assert isinstance(regressor.r2_lat, float)
#     assert isinstance(regressor.summary, dict)

# def test_regressor_assign_unknown_and_get_results():

#     data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
#                     sample_data="tests/test_data/testNA.txt")
#     regressor = PopRegressor(data_obj, output_folder=TEST_OUTPUT_FOLDER)
#     regressor.train()
#     regressor.test()
#     unknown_data = regressor.assign_unknown()

#     assert unknown_data.equals(regressor.regression)
#     assert os.path.exists(os.path.join(regressor.output_folder,
#                           "regressor_assignment_results.csv"))
#     os.remove(os.path.join(regressor.output_folder,
#                             "regressor_assignment_results.csv"))

#     assign_sum = regressor.get_assignment_summary(save=False)
#     assert isinstance(assign_sum, dict)
#     assert not os.path.exists(os.path.join(regressor.output_folder,
#                               "regressor_classification_summary.csv"))

#     with pytest.raises(ValueError, match=re.escape("Must run classify_by_contours() before getting summary.")):
#         regressor.get_classification_summary(save=False)

#     site_rank = regressor.rank_site_importance()
#     assert isinstance(site_rank, pd.DataFrame)
#     assert site_rank.empty == False
#     assert len(regressor.data.data.alleles[0]) == len(site_rank)

# def test_regressor_classify_by_contours():

#     data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
#                     sample_data="tests/test_data/testNA.txt")
#     regressor = PopRegressor(data_obj, output_folder=TEST_OUTPUT_FOLDER)

#     with pytest.raises(TypeError, match="nboots must be an integer"):
#         regressor.classify_by_contours(nboots=0.5)

#     with pytest.raises(TypeError, match="num_contours must be an integer"):
#         regressor.classify_by_contours(num_contours=0.5)

#     with pytest.raises(TypeError, match="save_plots must be a boolean"):
#         regressor.classify_by_contours(save_plots="True")

#     with pytest.raises(TypeError, match="save must be a boolean"):
#         regressor.classify_by_contours(save="True")

#     regressor.classify_by_contours()

#     assert isinstance(regressor.classification_test_results, pd.DataFrame)
#     assert regressor.classification_test_results.empty == False
#     assert isinstance(regressor.contour_classification, pd.DataFrame)
#     assert regressor.contour_classification.empty == False
#     assert isinstance(regressor.classification_confusion_matrix, np.ndarray)
#     assert isinstance(regressor.classification_accuracy, float)
#     assert isinstance(regressor.classification_precision, float)
#     assert isinstance(regressor.classification_recall, float)
#     assert isinstance(regressor.classification_f1, float)

#     class_sum = regressor.get_classification_summary(save=False)
#     assert isinstance(class_sum, dict)
#     assert class_sum["accuracy"] == regressor.classification_accuracy
#     assert class_sum["precision"] == regressor.classification_precision
#     assert class_sum["recall"] == regressor.classification_recall
#     assert class_sum["f1"] == regressor.classification_f1
#     assert (class_sum["confusion_matrix"] == regressor.classification_confusion_matrix).all()

# def test_regressor_save_and_load():

#     data_obj = GeneticData(genetic_data="tests/test_data/test.vcf", 
#                     sample_data="tests/test_data/testNA.txt")
#     regressor = PopRegressor(data_obj, output_folder=TEST_OUTPUT_FOLDER)
#     regressor.train()
#     regressor.test()
#     regressor.assign_unknown()
#     regressor.save()

#     assert os.path.exists(os.path.join(regressor.output_folder,
#                           "regressor.pkl"))

#     regressor2 = PopRegressor.load(load_path=os.path.join(regressor.output_folder,
#                                 "regressor.pkl"))

#     assert regressor2.train_history.equals(regressor.train_history)
#     assert regressor2.test_results.equals(regressor.test_results)
#     assert regressor2.median_distance == regressor.median_distance
#     assert regressor2.mean_distance == regressor.mean_distance
#     assert regressor2.r2_lat == regressor.r2_lat
#     assert regressor2.r2_long == regressor.r2_long
#     assert regressor2.regression.equals(regressor.regression)
#     assert isinstance(regressor2.best_model, RegressorNet)
#     assert isinstance(regressor.best_model, RegressorNet)

#     os.remove(os.path.join(regressor.output_folder,
#                             "regressor.pkl"))

#     regressor.classify_by_contours()
#     regressor.save()

#     regressor2 = PopRegressor.load(load_path=os.path.join(regressor.output_folder,
#                                 "regressor.pkl"))

#     assert regressor2.train_history.equals(regressor.train_history)
#     assert regressor2.test_results.equals(regressor.test_results)
#     assert regressor2.contour_classification.equals(regressor.contour_classification)
#     assert regressor2.classification_test_results.equals(regressor.classification_test_results)
#     assert regressor2.classification_accuracy == regressor.classification_accuracy
#     assert regressor2.classification_precision == regressor.classification_precision
#     assert regressor2.classification_recall == regressor.classification_recall
#     assert regressor2.classification_f1 == regressor.classification_f1
#     assert regressor2.classification_confusion_matrix.tolist() == regressor.classification_confusion_matrix.tolist()
#     assert isinstance(regressor2.best_model, RegressorNet)
#     assert isinstance(regressor.best_model, RegressorNet)

#     os.remove(os.path.join(regressor.output_folder,
#                             "regressor.pkl"))


import pandas as pd
import multiprocessing as mp
import argparse
import os

from classifier import PopClassifier

def _test_on_bootstraps(rep, boot, popfinder_path):
    
    classifier_object = PopClassifier.load(os.path.join(
        popfinder_path, f"rep{rep}_boot{boot}", "classifier.pkl"))
    
    # Test for all CV splits
    classifier_object.test(use_best_model=False)

    # Return losses
    return classifier_object.test_results

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", help="Path to PopClassifier object")
    parser.add_argument("-n", help="Number of bootstraps", type=int)
    parser.add_argument("-r", help="Number of repetitions", type=int)
    parser.add_argument("-j", help="Number of jobs", type=int)
    args = parser.parse_args()
    popfinder_path = args.p
    nboots = args.n
    nreps = args.r
    num_jobs = args.j

    if num_jobs == -1:
        num_jobs = mp.cpu_count()
    pool = mp.Pool(processes=num_jobs)
    results = pool.starmap(
        _test_on_bootstraps, 
        [(rep, boot, popfinder_path) for rep, boot in zip(range(nreps), range(nboots))])
    pool.close()
    pool.join()

    for rep in range(nreps):
        for boot in range(nboots):
            ind = rep * nboots + boot
            results[ind]["rep"] = rep + 1
            results[ind]["bootstrap"] = boot + 1
    
    final_results = pd.concat(results, ignore_index=True)
    final_results.to_csv(os.path.join(popfinder_path, "test_results.csv"), index=False)


import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.autograd import Variable
from sklearn import preprocessing
import numpy as np
import dill
import os
import sys

def _generate_train_inputs(data_obj, valid_size, cv_splits, cv_reps, seed=123, bootstrap=False):

    if cv_splits == 1:
        train_input, valid_input = data_obj.split_train_test(
            data_obj.train, test_size=valid_size, seed=seed, bootstrap=bootstrap)
        inputs = [(train_input, valid_input)]

    elif cv_splits > 1:
        inputs = data_obj.split_kfcv(
            data_obj.train, n_splits=cv_splits, n_reps=cv_reps, seed=seed, bootstrap=bootstrap)

    return inputs

def _split_input_classifier(clf, input):
        
    train_input, valid_input = input

    X_train = train_input["alleles"]
    X_valid = valid_input["alleles"]
    y_train = train_input["pop"] # one hot encode
    y_valid = valid_input["pop"] # one hot encode

    # Label encode y values
    # clf.label_enc = preprocessing.LabelEncoder()
    y_train = clf.label_enc.fit_transform(y_train)
    y_valid = clf.label_enc.transform(y_valid)

    X_train, y_train = _data_converter(X_train, y_train)
    X_valid, y_valid = _data_converter(X_valid, y_valid)

    return X_train, y_train, X_valid, y_valid

def _split_input_regressor(input):
        
    train_input, valid_input = input

    X_train = train_input["alleles"]
    X_valid = valid_input["alleles"]
    y_train = train_input[["x", "y"]]
    y_valid = valid_input[["x", "y"]]

    X_train, y_train = _data_converter(X_train, y_train)
    X_valid, y_valid = _data_converter(X_valid, y_valid)

    return X_train, y_train, X_valid, y_valid

def _generate_data_loaders(X_train, y_train, X_valid, y_valid, batch_size=16):

    train = TensorDataset(X_train, y_train)
    train_loader = DataLoader(train, batch_size=batch_size, shuffle=True, drop_last=True)
    valid = TensorDataset(X_valid, y_valid)
    valid_loader = DataLoader(valid, batch_size=len(valid.tensors[0]), shuffle=True, drop_last=True)

    return train_loader, valid_loader

def _data_converter(x, y, variable=False):

    features = torch.from_numpy(np.vstack(np.array(x)).astype(np.float32))
    if torch.isnan(features).sum() != 0:
        print("Remove NaNs from features")        
    if variable:
        features = Variable(features)

    if y is not None:
        targets = torch.from_numpy(np.vstack(np.array(y)))
        if torch.isnan(targets).sum() != 0:
            print("remove NaNs from target")
        if variable:
            targets = Variable(targets)
            
        return features, targets

    else:
        return features

def _save(obj, save_path=None, file="model.pkl"):
    """
    Saves the current instance of the class to a pickle file.
    """
    if save_path is None:
        save_path = obj.output_folder
    with open(os.path.join(save_path, file), "wb") as f:
        dill.dump(obj, f)

def _load(load_path=None):
    """
    Loads a saved instance of the class from a pickle file.
    """
    sys.path.append(os.path.dirname(__file__))
    with open(load_path, "rb") as f:
        return dill.load(f)


import argparse
from popfinder.dataloader import GeneticData
from popfinder.regressor import PopRegressor

def main():

    parser = argparse.ArgumentParser(
        prog="pop_regressor",
        description="PopRegressor: A tool for population assignment from" +
        " genetic data using regression neural networks."
    )

    # Arguments for determining which function to use
    parser.add_argument('--load_data', action='store_true', help='Load data from genetic and sample data files.')
    parser.add_argument('--train', action='store_true', help='Train a neural network.')
    parser.add_argument('--test', action='store_true', help='Test a trained neural network.')
    parser.add_argument('--assign', action='store_true', help='Assign samples to populations.')
    parser.add_argument('--rank_site_importance', action='store_true', help='Rank sites by importance for regression.')
    parser.add_argument('--plot_training_curve', action='store_true', help='Plot the training curve.')
    parser.add_argument('--plot_location', action='store_true', help='Plot the xy coordinates of samples compared to population centroids.')
    parser.add_argument('--classify_by_contours', action='store_true', help='Classify samples using regression neural network results kernel density estimates.')
    parser.add_argument('--plot_contour_map', action='store_true', help='Plot the contour maps of the predicted sample locations.')
    parser.add_argument('--plot_confusion_matrix', action='store_true', help='Plot the confusion matrix of the classification results.')
    parser.add_argument('--plot_structure', action='store_true', help='Plot the structure of population assignment.')
    parser.add_argument('--plot_assignment', action='store_true', help='Plot the assignment of samples to populations.')
    parser.add_argument('--get_assignment_summary', action='store_true', help='Get a summary of the assignment results.')
    parser.add_argument('--get_classification_summary', action='store_true', help='Get a summary of the classification results.')

    # Arguments for loading data
    parser.add_argument('--genetic_data', type=str, default=None)
    parser.add_argument('--sample_data', type=str, default=None)
    parser.add_argument('--test_size', type=float, default=None)
    parser.add_argument('--seed', type=int, default=None)
    parser.add_argument('--output_folder', type=str, default=None)

    # Arguments for training
    parser.add_argument('--epochs', type=int, default=None)
    parser.add_argument('--valid_size', type=float, default=None)
    parser.add_argument('--cv_splits', type=int, default=None)
    parser.add_argument('--cv_reps', type=int, default=None)

    # Arguments for contour classification
    parser.add_argument('--nboots', type=int, default=None)
    parser.add_argument('--num_contours', type=int, default=None)

    # Arguments for plotting structure/assignment
    parser.add_argument('--col_scheme', type=str, default=None)

    args = parser.parse_args()

    if args.load_data:
        data = GeneticData(args.genetic_data, args.sample_data, args.test_size, args.seed)
        regressor = PopRegressor(data, args.output_folder)
        regressor.save()

    elif args.train:
        regressor = PopRegressor.load(args.output_folder)
        regressor.train(args.epochs, args.valid_size, args.cv_splits, args.cv_reps)
        regressor.save()
    
    elif args.test:
        regressor = PopRegressor.load(args.output_folder)
        regressor.test()
        regressor.save()

    elif args.assign:
        regressor = PopRegressor.load(args.output_folder)
        regressor.assign_unknown()
        regressor.save()

    elif args.rank_site_importance:
        regressor = PopRegressor.load(args.output_folder)
        regressor.rank_site_importance()
        regressor.save()

    elif args.plot_training_curve:
        regressor = PopRegressor.load(args.output_folder)
        regressor.plot_training_curve()

    elif args.plot_location:
        regressor = PopRegressor.load(args.output_folder)
        regressor.plot_location()

    elif args.classify_by_contours:
        regressor = PopRegressor.load(args.output_folder)
        regressor.classify_by_contours(args.nboots, args.num_contours)
        regressor.save()

    elif args.plot_contour_map:
        regressor = PopRegressor.load(args.output_folder)
        regressor.plot_contour_map()

    elif args.plot_confusion_matrix:
        regressor = PopRegressor.load(args.output_folder)
        regressor.plot_confusion_matrix()

    elif args.plot_structure:
        regressor = PopRegressor.load(args.output_folder)
        regressor.plot_structure(args.col_scheme)

    elif args.plot_assignment:
        regressor = PopRegressor.load(args.output_folder)
        regressor.plot_assignment(args.col_scheme)

    elif args.get_assignment_summary:
        regressor = PopRegressor.load(args.output_folder)
        regressor.get_assignment_summary()

    elif args.get_classification_summary:
        regressor = PopRegressor.load(args.output_folder)
        regressor.get_classification_summary()        

    else:
        print("No function selected. Use --help for more information.")

if __name__ == "__main__":
    main()


import torch.nn as nn
import torch.nn.functional as F
import torch

class ClassifierNet(nn.Module):
    
    def __init__(self, input_size, hidden_size, output_size,
                 batch_size, dropout_prop):
        super(ClassifierNet, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.batch1 = nn.BatchNorm1d(hidden_size)
        self.fc2 = nn.Linear(hidden_size, batch_size)
        self.batch2 = nn.BatchNorm1d(batch_size)
        self.fc3 = nn.Linear(batch_size, output_size)
        self.dropout = nn.Dropout(dropout_prop)

    def forward(self, x):
        x = self.batch1(self.dropout(F.relu(self.fc1(x))))
        x = self.batch2(self.dropout(F.relu(self.fc2(x))))
        x = F.softmax(self.fc3(x), dim=1)

        return x

class RegressorNet(nn.Module):

    def __init__(self, input_size, hidden_size, batch_size, dropout_prop):
        super(RegressorNet, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.batch1 = nn.BatchNorm1d(hidden_size)
        self.fc2 = nn.Linear(hidden_size, batch_size)
        self.batch2 = nn.BatchNorm1d(batch_size)
        self.fc3 = nn.Linear(batch_size, batch_size)
        self.fc4 = nn.Linear(batch_size, 2)
        self.fc5 = nn.Linear(2, 2)
        self.dropout = nn.Dropout(dropout_prop)

    def forward(self, x):
        x = self.batch1(self.dropout(F.elu(self.fc1(x))))
        x = self.batch2(self.dropout(F.elu(self.fc2(x))))
        x = F.elu(self.fc3(x))
        x = F.elu(self.fc3(x))
        x = self.fc5(self.fc4(x))

        return x


import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import pandas as pd
import numpy as np
import itertools
import os

def _plot_training_curve(train_history, nn_type, output_folder, save, facet_by_split_rep):

    plot_data = train_history.rename(columns={"valid": "validation"})
    plot_data = pd.melt(plot_data, id_vars=["epoch", "split", "rep"], 
            value_vars=["train", "validation"], 
            var_name="dataset", value_name="loss")

    d = {'color': ['C0', 'k'], "ls" : ["-","--"]}

    if facet_by_split_rep:
        g = sns.FacetGrid(plot_data, col="split", row="rep", 
                          hue="dataset", hue_kws=d, height=2, aspect=1.5)
    else:
        g = sns.FacetGrid(plot_data, hue="dataset", hue_kws=d, height=2, aspect=1.5)
    
    g.map(sns.lineplot, "epoch", "loss", lw=0.5)
    g.add_legend(title="")

    if save:
        g.savefig(os.path.join(output_folder, nn_type + "_training_curve.png"),
            bbox_inches="tight")

def _plot_confusion_matrix(test_results, confusion_matrix, nn_type,
    output_folder, save):

    true_labels = test_results["true_pop"] # test w/ classifier

    cm = np.round(confusion_matrix, 2)
    plt.style.use("default")
    plt.figure()
    plt.imshow(cm, cmap="Blues")
    plt.colorbar()
    plt.ylabel("True Population")
    plt.xlabel("Predicted Population")
    plt.title("Confusion Matrix")
    tick_marks = np.arange(len(np.unique(true_labels)))
    plt.xticks(tick_marks, np.unique(true_labels))
    plt.yticks(tick_marks, np.unique(true_labels))
    thresh = cm.max() / 2.0
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, cm[i, j], horizontalalignment="center",
            color="white" if cm[i, j] > thresh else "black")
    plt.tight_layout()

    if save:
        plt.savefig(os.path.join(output_folder, nn_type + "_confusion_matrix.png"))

def _plot_assignment(e_preds, col_scheme, output_folder,
    nn_type, save, best_model_only):
        
    e_preds.set_index("sampleID", inplace=True)

    if nn_type == "regressor":
        e_preds = pd.get_dummies(e_preds["classification"], dtype=float)
    elif nn_type == "classifier":
        e_preds = pd.get_dummies(e_preds["assigned_pop"], dtype=float)

    if not best_model_only:
        e_preds = e_preds.reset_index().groupby("sampleID").mean()

    num_classes = len(e_preds.columns)

    sns.set()
    sns.set_style("ticks")
    e_preds.plot(kind="bar", stacked=True,
        colormap=ListedColormap(sns.color_palette(col_scheme, num_classes)),
        figsize=(12, 6), grid=None)
    legend = plt.legend(
        loc="center right",
        bbox_to_anchor=(1.2, 0.5),
        prop={"size": 15},
        title="Predicted Population",
    )
    plt.setp(legend.get_title(), fontsize="x-large")
    plt.xlabel("Sample ID", fontsize=20)
    plt.ylabel("Frequency of Assignment", fontsize=20)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.tight_layout()

    if save:
        plt.savefig(os.path.join(
            output_folder, nn_type + "_assignment.png"),
            bbox_inches="tight")

def _plot_structure(preds, col_scheme, nn_type, output_folder, save):

    num_classes = len(preds.index)

    sns.set()
    sns.set_style("ticks")
    preds.plot(kind="bar", stacked=True,
        colormap=ListedColormap(sns.color_palette(col_scheme, num_classes)),
        figsize=(12, 6), grid=None)
    legend = plt.legend(loc="center right", bbox_to_anchor=(1.2, 0.5),
        prop={"size": 15}, title="Predicted Pop")
    plt.setp(legend.get_title(), fontsize="x-large")
    plt.xlabel("Actual Pop", fontsize=20)
    plt.ylabel("Frequency of Assignment", fontsize=20)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    if save:
        plt.savefig(os.path.join(output_folder,
            nn_type + "_structure.png"),
            bbox_inches="tight")


import argparse
import os
from popfinder.dataloader import GeneticData
from popfinder.classifier import PopClassifier

def load_classifier(arg_output_folder):

    classifier_name = "classifier.pkl"
    output_folder = arg_output_folder if arg_output_folder is not None else os.getcwd()
    classifier = PopClassifier.load(os.path.join(output_folder, classifier_name))

    return classifier


def main():

    parser = argparse.ArgumentParser(
        prog="pop_classifier",
        description="PopClassifier: A tool for population assignment from" + 
        " genetic data using classification neural networks."
    )

    # Arguments for determining which function to use
    parser.add_argument('--load_data', action='store_true', help='Load data from genetic and sample data files.')
    parser.add_argument('--train', action='store_true', help='Train a neural network.')
    parser.add_argument('--test', action='store_true', help='Test a trained neural network.')
    parser.add_argument('--assign', action='store_true', help='Assign samples to populations.')
    parser.add_argument('--rank_site_importance', action='store_true', help='Rank sites by importance for classification.')
    parser.add_argument('--plot_training_curve', action='store_true', help='Plot training curve.')
    parser.add_argument('--plot_confusion_matrix', action='store_true', help='Plot confusion matrix.')
    parser.add_argument('--plot_structure', action='store_true', help='Plot structure of population assignment.')
    parser.add_argument('--plot_assignment', action='store_true', help='Plot assignment of samples to populations.')

    # Arguments for loading data
    parser.add_argument('--genetic_data', type=str, default=None, help='Path to genetic data file.')
    parser.add_argument('--sample_data', type=str, default=None, help='Path to sample data file.')
    parser.add_argument('--test_size', type=float, default=0.2, help='Proportion of samples to use for testing.')
    parser.add_argument('--seed', type=int, default=123, help='Random seed for splitting data into training and testing sets.')
    parser.add_argument('--output_folder', type=str, default=None, help='Path to output folder. (Default: current directory)')

    # Arguments for training
    parser.add_argument('--epochs', type=int, default=100, help='Number of epochs to train for.')
    parser.add_argument('--valid_size', type=float, default=0.2, help='Proportion of training data to use for validation.')
    parser.add_argument('--cv_splits', type=int, default=1, help='Number of cross-validation splits.')
    parser.add_argument('--cv_reps', type=int, default=1, help='Number of cross-validation repetitions.')

    # Arguments for plotting structure/assignment
    parser.add_argument('--col_scheme', type=str, default="Spectral", help='Color scheme for plotting structure/assignment. (Default: "Spectral")')

    args = parser.parse_args()

    if args.load_data:
        data = GeneticData(args.genetic_data, args.sample_data, args.test_size, args.seed)
        classifier = PopClassifier(data, output_folder=args.output_folder)
        classifier.save()

    elif args.train:
        classifier = load_classifier(args.output_folder)
        classifier.train(args.epochs, args.valid_size, args.cv_splits, args.cv_reps)
        classifier.save()

    elif args.test:
        classifier = load_classifier(args.output_folder)
        classifier.test()
        classifier.get_classification_summary(save=True)
        classifier.save()

    elif args.assign:
        classifier = load_classifier(args.output_folder)
        classifier.assign_unknown(save=True)
        classifier.save()

    elif args.rank_site_importance:
        classifier = load_classifier(args.output_folder)
        classifier.rank_site_importance()
        classifier.save()

    elif args.plot_training_curve:
        classifier = load_classifier(args.output_folder)
        classifier.plot_training_curve()

    elif args.plot_confusion_matrix:
        classifier = load_classifier(args.output_folder)
        classifier.plot_confusion_matrix()

    elif args.plot_structure:
        classifier = load_classifier(args.output_folder)
        classifier.plot_structure(args.col_scheme)

    elif args.plot_assignment:
        classifier = load_classifier(args.output_folder)
        classifier.plot_assignment(args.col_scheme)

    else:         
        print("No function selected. Use --help for more information")
    

if __name__ == "__main__":
    main()
    



import torch
from torch.autograd import Variable
from torch.optim.lr_scheduler import ReduceLROnPlateau
from scipy import spatial
from scipy import stats
from sklearn.metrics import accuracy_score, confusion_matrix, f1_score, precision_score, recall_score
from subprocess import call
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import matplotlib
import seaborn as sns
import multiprocessing as mp
import numpy as np
import pandas as pd
import os
import tempfile


from popfinder.dataloader import GeneticData
from popfinder._neural_networks import RegressorNet
from popfinder._helper import _generate_train_inputs
from popfinder._helper import _generate_data_loaders
from popfinder._helper import _data_converter
from popfinder._helper import _split_input_regressor
from popfinder._helper import _save, _load
from popfinder._visualize import _plot_assignment
from popfinder._visualize import _plot_training_curve
from popfinder._visualize import _plot_confusion_matrix
from popfinder._visualize import _plot_structure
import popfinder as pf

import warnings
warnings.filterwarnings("ignore")

class PopRegressor(object):
    """
    A class to represent a regressor neural network object for population assignment.

    Parameters
    ----------
    data : GeneticData object
        GeneticData object containing the training data.
    nboots : int, optional
        Number of bootstrap samples to generate. The default is 20.
    random_state : int, optional
        Random seed for reproducibility. The default is 123.
    output_folder : str, optional
        Path to output folder. The default is None.
    
    Attributes
    ----------
    data : GeneticData object
        GeneticData object containing the training data.
    nboots : int
        Number of bootstrap samples to generate.
    random_state : int
        Random seed for reproducibility.
    output_folder : str
        Path to output folder.
    train_history : list
        List of training history objects.
    best_model : torch.nn.Module
        Best model from training.
    regression : pandas.DataFrame
        Dataframe containing the regression results.
    median_distance : float
        Median distance between true and predicted coordinates.
    mean_distance : float
        Mean distance between true and predicted coordinates.
    r2_lat : float
        R-squared value for latitude.
    r2_long : float
        R-squared value for longitude.
    summary : pandas.DataFrame
        Dataframe containing the summary of the regression results.
    contour_classification : pandas.DataFrame
        Dataframe containing the contour classification results.
    classification_test_results : pandas.DataFrame
        Dataframe containing the classification test results.
    classification_accuracy : float
        Classification accuracy.
    classification_precision : float
        Classification precision.
    classification_recall : float
        Classification recall.
    classification_f1 : float
        Classification F1 score.
    classification_confusion_matrix : numpy.ndarray
        Classification confusion matrix.

    Methods
    -------
    train(epochs=100, valid_size=0.2, cv_splits=1, cv_reps=1, learning_rate=0.001, batch_size=16, dropout_prop=0)
        Train the regressor.
    test()
        Test the regressor.
    assign_unknown()
        Assign unknown samples.
    rank_site_importance()
        Rank the importance of each site.
    plot_training_curve()
        Plot the training curve.
    plot_location()
        Plot the predicted and true locations.  
    plot_contour_map()
        Plot the contour map.
    plot_confusion_matrix()
        Plot the confusion matrix.
    plot_structure()
        Plot the neural network structure.  
    save()
        Save the regressor object.
    load()
        Load the regressor object.  
    """
    def __init__(self, data, nboots=20, random_state=123, output_folder=None):

        self._validate_init_inputs(data, nboots, random_state, output_folder)

        self.__data = data # GeneticData object
        self.__nboots = nboots
        self.__random_state = random_state
        if output_folder is None:
            output_folder = os.getcwd()
        self.__output_folder = output_folder
        self.__boot_data = None
        self.__train_history = None
        self.__best_model = None
        self.__test_results = None
        self.__regression = None
        self.__median_distance = None
        self.__mean_distance = None
        self.__r2_lat = None
        self.__r2_long = None
        self.__summary = None
        self.__contour_classification = None
        self.__classification_test_results = None
        self.__classification_accuracy = None
        self.__classification_precision = None
        self.__classification_recall = None
        self.__classification_f1 = None
        self.__classification_confusion_matrix = None
        self.__nn_type = "regressor"
        self.__lowest_val_loss = 9999

    @property
    def data(self):
        return self.__data

    @property
    def nboots(self):
        return self.__nboots

    @property
    def random_state(self):
        return self.__random_state

    @property
    def output_folder(self):
        return self.__output_folder

    @output_folder.setter
    def output_folder(self, output_folder):
        self.__output_folder = output_folder

    @property
    def boot_data(self):
        return self.__boot_data
    
    @boot_data.setter
    def boot_data(self, boot_data):
        self.__boot_data = boot_data

    @property
    def train_history(self):
        return self.__train_history

    @property
    def best_model(self):
        return self.__best_model

    @property
    def test_results(self):
        return self.__test_results

    @property
    def regression(self):
        return self.__regression

    @property
    def median_distance(self):
        return self.__median_distance

    @property
    def mean_distance(self):
        return self.__mean_distance

    @property
    def r2_lat(self):
        return self.__r2_lat

    @property
    def r2_long(self):
        return self.__r2_long

    @property
    def summary(self):
        return self.__summary

    @property
    def contour_classification(self):
        return self.__contour_classification

    @property
    def classification_test_results(self):
        return self.__classification_test_results

    @property
    def classification_accuracy(self):
        return self.__classification_accuracy

    @property
    def classification_precision(self): 
        return self.__classification_precision

    @property
    def classification_recall(self):
        return self.__classification_recall

    @property
    def classification_f1(self):
        return self.__classification_f1

    @property
    def classification_confusion_matrix(self):
        return self.__classification_confusion_matrix

    @property
    def nn_type(self):
        return self.__nn_type

    @property
    def lowest_val_loss(self):
        return self.__lowest_val_loss

    def train(self, epochs=100, valid_size=0.2, cv_splits=1, cv_reps=1,
              learning_rate=0.001, batch_size=16, dropout_prop=0):
        """
        Trains the regression neural network to estimate xy coordinates of 
        a sample's origin.
        
        Parameters
        ----------
        epochs : int, optional
            Number of epochs to train the neural network. The default is 100.
        valid_size : float, optional
            Proportion of data to use for validation. The default is 0.2.
        cv_splits : int, optional
            Number of cross-validation splits. The default is 1.
        cv_reps : int, optional
            Number of cross-validation repetitions. The default is 1.
        learning_rate : float, optional
            Learning rate for the neural network. The default is 0.001.
        batch_size : int, optional
            Batch size for the neural network. The default is 16.
        dropout_prop : float, optional
            Dropout proportion for the neural network. The default is 0.
            
        Returns
        -------
        None.
        """
        self._validate_train_inputs(epochs, valid_size, cv_splits, cv_reps,
                            learning_rate, batch_size, dropout_prop)

        if self.__boot_data is None:
            inputs = _generate_train_inputs(self.data, valid_size, cv_splits,
                                            cv_reps, seed=self.random_state)
        else:
            inputs = _generate_train_inputs(self.__boot_data, valid_size, cv_splits,
                                            cv_reps, seed=self.random_state)

        loss_df_final = pd.DataFrame({"rep": [], "split": [], "epoch": [],
                                      "train": [], "valid": []})
        lowest_val_loss = 9999

        for i, input in enumerate(inputs):

            X_train, y_train, X_valid, y_valid = _split_input_regressor(input)
            net = RegressorNet(input_size=X_train.shape[1], hidden_size=32,
                               batch_size=batch_size, dropout_prop=dropout_prop)
            optimizer = torch.optim.Adam(net.parameters(), lr=learning_rate)
            scheduler = ReduceLROnPlateau(optimizer, factor=0.5)
            # loss_func = self._euclidean_dist_loss
            loss_func = torch.nn.MSELoss()

            y_train = torch.tensor(self._normalize_locations(y_train))
            y_valid = torch.tensor(self._normalize_locations(y_valid))

            train_loader, valid_loader = _generate_data_loaders(X_train, y_train,
                                                                X_valid, y_valid)
            loss_dict = {"epoch": [], "train": [], "valid": []}

            for epoch in range(epochs):

                train_loss = 0
                valid_loss = 0

                for _, (data, target) in enumerate(train_loader):
                    optimizer.zero_grad()
                    output = net(data)
                    loss = loss_func(output.squeeze(), target.squeeze().float())
                    loss.backward()
                    optimizer.step()
                    train_loss += loss.data.item()
            
                # Calculate average train loss
                avg_train_loss = train_loss / len(train_loader)

                for _, (data, target) in enumerate(valid_loader):
                    output = net(data)
                    loss = loss_func(output.squeeze(), target.squeeze().long())
                    valid_loss += loss.data.item()

                    if valid_loss < lowest_val_loss:
                        lowest_val_loss = valid_loss
                        torch.save(net, os.path.join(self.output_folder, "best_model.pt"))

                # Step scheduler for LR decay
                scheduler.step(valid_loss)

                # Calculate average validation loss
                avg_valid_loss = valid_loss / len(valid_loader)

                loss_dict["epoch"].append(epoch)
                loss_dict["train"].append(avg_train_loss)
                loss_dict["valid"].append(avg_valid_loss)

            loss_df = pd.DataFrame(loss_dict)

            # loss_df = self._fit_regressor_model(epochs, train_loader,
            #         valid_loader, net, optimizer, scheduler, loss_func)

            split = i % cv_splits + 1
            rep = int(i / cv_splits) + 1

            loss_df["rep"] = rep
            loss_df["split"] = split

            loss_df_final = pd.concat([loss_df_final, loss_df])

        self.__train_history = loss_df_final
        self.__best_model = torch.load(os.path.join(self.output_folder, "best_model.pt"))

    def test(self, save=True, verbose=False):
        """
        Tests the regression neural network on the test data.
        
        Parameters
        ----------
        verbose : bool, optional
            Whether to print the test results. The default is False.

        Returns
        -------
        None.
        """
        
        if self.__boot_data is None:
            test_input = self.data.test
        else:
            test_input = self.__boot_data.test

        X_test = test_input["alleles"]    
        y_test = test_input[["x", "y"]]
        X_test, y_test = _data_converter(X_test, y_test)
        y_test = self._unnormalize_locations(y_test)

        y_pred = self.best_model(X_test).detach().numpy()
        y_pred = self._unnormalize_locations(y_pred)

        test_input = test_input.assign(x_pred=y_pred[:, 0], y_pred=y_pred[:, 1])

        dists = [
            spatial.distance.euclidean(
                y_pred[x, :], y_test[x, :]
            ) for x in range(len(y_pred))
        ]

        # true coordinates vs predicted coordinates
        self.__test_results = test_input

        if save:
            self.test_results.to_csv(os.path.join(self.output_folder,
                                     "regressor_test_results.csv"), index=False)

        self.__median_distance = np.median(dists)
        self.__mean_distance = np.mean(dists)
        self.__r2_long = np.corrcoef(y_pred[:, 0], y_test[:, 0])[0][1] ** 2
        self.__r2_lat = np.corrcoef(y_pred[:, 1], y_test[:, 1])[0][1] ** 2

        self.__summary = self.get_assignment_summary()

        if verbose:
            print(self.summary)


    def assign_unknown(self, save=True):
        """
        Assigns unknown samples to their predicted origin.
        
        Parameters
        ----------
        save : bool, optional
            Whether to save the results to a csv file. The default is True.
        
        Returns
        -------
        None.
        """
        
        if self.__boot_data is None:
            unknown_data = self.data.unknowns
        else:
            unknown_data = self.__boot_data.unknowns

        X_unknown = unknown_data["alleles"]
        X_unknown = _data_converter(X_unknown, None)

        y_pred = self.best_model(X_unknown).detach().numpy()
        y_pred = self._unnormalize_locations(y_pred)
        unknown_data.loc[:, "x_pred"] = y_pred[:, 0]
        unknown_data.loc[:, "y_pred"] = y_pred[:, 1]

        self.__regression = unknown_data

        if save:
            unknown_data.to_csv(os.path.join(self.output_folder,
                                "regressor_assignment_results.csv"))

        return unknown_data
    
    def perform_bootstrap_regression(self, nboots=5, nreps=5,
        epochs=100, valid_size=0.2, cv_splits=1, cv_reps=1,
        learning_rate=0.001, batch_size=16, dropout_prop=0, 
        jobs=-1, save_plots=True, save=True):
        """
        Generates many predictions using bootstraps of the original data.

        Parameters
        ----------
        nboots : int, optional
            Number of bootstraps to perform. The default is 5.
        nreps : int, optional
            Number of repetitions for each bootstrap. The default is 5.
        save_plots : bool, optional
            Whether to save the contour plots. The default is True.
        save : bool, optional
            Whether to save the classification results. The default is True.
        
        Returns
        -------
        None.
        """
        self._generate_bootstrap_results(nboots, nreps, epochs, valid_size,
                                  cv_splits, cv_reps, learning_rate, batch_size,
                                  dropout_prop, jobs) 

        test_locs = self.test_locs_final
        pred_locs = self.pred_locs_final

    def classify_by_contours(self, nboots=5, nreps=5, num_contours=5,
        epochs=100, valid_size=0.2, cv_splits=1, cv_reps=1,
        learning_rate=0.001, batch_size=16, dropout_prop=0, 
        jobs=-1, save_plots=True, save=True):
        """
        Classifies unknown samples by kernel density estimates (contours).
        This function uses 2D kernel density estimation to create contour 
        plots from many predictions on the same sample, then assigns the sample
        to the contour with the highest probability.

        Parameters
        ----------
        nboots : int, optional
            Number of bootstraps to perform. The default is 5.
        nreps : int, optional
            Number of repetitions for each bootstrap. The default is 5.
        num_contours : int, optional
            Number of contours to generate. The default is 5.
        save_plots : bool, optional
            Whether to save the contour plots. The default is True.
        save : bool, optional
            Whether to save the classification results. The default is True.
        
        Returns
        -------
        None.
        """
        self._validate_contour_inputs(nboots, num_contours, save_plots, save)

        self._generate_bootstrap_results(nboots, nreps, epochs, valid_size,
                                  cv_splits, cv_reps, learning_rate, batch_size,
                                  dropout_prop, jobs) 

        test_locs = self.test_locs_final
        pred_locs = self.pred_locs_final

        self.__classification_test_results = self._test_classification(test_locs,
            num_contours, save_plots)
        self.__contour_classification = self._classify_unknowns(pred_locs,
            test_locs, num_contours, save_plots)

        if save:
            self.classification_test_results.to_csv(os.path.join(self.output_folder,
                "contour_classification_test_report.csv"), index=False)
            self.contour_classification.to_csv(os.path.join(self.output_folder,
                "contour_classification_results.csv"), index=False)

        # Generate classification summary stats from test_report
        y_pred = self.classification_test_results["pred_pop"]
        y_true = self.classification_test_results["true_pop"]
        self.__classification_confusion_matrix = np.round(
            confusion_matrix(y_true, y_pred, normalize="true"), 3)
        self.__classification_accuracy = accuracy_score(y_true, y_pred)
        self.__classification_precision = precision_score(y_true, y_pred, average="weighted")
        self.__classification_recall = recall_score(y_true, y_pred, average="weighted")
        self.__classification_f1 = f1_score(y_true, y_pred, average="weighted")

    # Reporting functions below
    def get_assignment_summary(self, save=True):
        """
        Returns a summary of the assignment results.
        
        Parameters
        ----------
        save : bool, optional
            Whether to save the summary to a csv file. The default is True.
            
        Returns
        -------
        dict
            Dictionary containing the summary statistics median distance,
            mean distance, r2_long, r2_lat.
        """
        summary = {
            "median_distance": [self.median_distance],
            "mean_distance": [self.mean_distance],
            "r2_long": [self.r2_long],
            "r2_lat": [self.r2_lat]
        }

        if save:
            pd.DataFrame(summary).to_csv(os.path.join(self.output_folder,
                "regressor_assignment_summary.csv"), index=False)

        return summary

    def get_classification_summary(self, save=True):
        """
        Returns a summary of the classification results.
        
        Parameters
        ----------
        save : bool, optional
            Whether to save the summary to a csv file. The default is True.
        
        Returns
        -------
        dict
            Dictionary containing the summary statistics accuracy, precision,
            recall, f1, and confusion matrix.
        """
        if self.classification_test_results is None:
            raise ValueError("Must run classify_by_contours() before getting summary.")

        summary = { # need to grab all these items
            "accuracy": [self.classification_accuracy],
            "precision": [self.classification_precision],
            "recall": [self.classification_recall],
            "f1": [self.classification_f1],
            "confusion_matrix": [self.classification_confusion_matrix]
        }
        
        if save:
            pd.DataFrame(summary).to_csv(os.path.join(self.output_folder,
                "regressor_classification_summary.csv"), index=False)

        return summary

    def rank_site_importance(self, save=True):
        """
        Rank sites (SNPs) by importance in model performance.

        Parameters
        ----------
        save : bool, optional
            Whether to save the results to a csv file. The default is True.
        
        Returns
        -------
        pandas.DataFrame
            Dataframe containing the ranked SNPs and their importance scores.
        """
        if self.best_model is None:
            raise ValueError("Model has not been trained yet. " + 
            "Please run the train() method first.")

        X = self.data.knowns["alleles"].to_numpy()
        X = np.stack(X)
        Y = self.data.knowns[["x", "y"]].to_numpy()
        snp_names = np.arange(1, X.shape[1] + 1)
        errors = []

        for i in range(X.shape[1]):
            X_temp = X.copy()
            X_temp[:, i] = np.random.choice(X_temp[:, i], X_temp.shape[0])
            X_temp = torch.from_numpy(X_temp).float()
            preds = self.best_model(X_temp).detach().numpy()
            preds = self._unnormalize_locations(preds)
            errors.append(spatial.distance.cdist(preds, Y).mean())
       
        max_error = np.max(errors)
        importance = [1 - (e / max_error) for e in errors]
        importance_data = {"snp": snp_names, "error": errors,
                           "importance": importance}
        ranking = pd.DataFrame(importance_data).sort_values("importance",
                                                            ascending=False)

        if save:
            ranking.to_csv(os.path.join(self.output_folder,
                "regressor_site_importance.csv"), index=False)

        return ranking

    # Plotting functions below
    def plot_training_curve(self, save=True):
        """
        Plot the training curve of the model.
        
        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
            
        Returns
        -------
        None.
        """

        _plot_training_curve(self.train_history, self.nn_type,
            self.output_folder, save)

    def plot_location(self, sampleID=None, save=True):
        """
        Plot the predicted location of a sample of unknown origin
        compared to known locations of populations.
        
        Parameters
        ----------
        sampleID : str, optional
            The sampleID to plot. If None, all samples will be plotted.
            The default is None.
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
            
        Returns
        -------
        None.
        """
        if self.regression is None:
            raise Exception("No regression data available. Please run" + 
                " assign_unknown() first.")

        if sampleID is None:
            sample_list = self.data.unknowns.sampleID.unique()
        else:
            sample_list = [sampleID]

        unique_test_data = self.data.test.drop_duplicates(subset=["pop"])

        sample_data = pd.DataFrame()
        for sample in sample_list:
            pred_sample_df = pd.DataFrame()
            pred_sample_df["x"] = self.regression[self.regression["sampleID"] == sample
                ]["x_pred"].values
            pred_sample_df["y"] = self.regression[self.regression["sampleID"] == sample
                ]["y_pred"].values
            pred_sample_df["pop"] = "prediction"
            pred_sample_df["sampleID"] = sample

            pop_sample_df = pd.DataFrame()
            pop_sample_df["x"] = unique_test_data["x"].values
            pop_sample_df["y"] = unique_test_data["y"].values
            pop_sample_df["pop"] = unique_test_data["pop"].values
            pop_sample_df["sampleID"] = sample

            sample_data = pd.concat([sample_data, pred_sample_df, pop_sample_df], axis=0)

        # TODO: Create custom colour palette to highlight pred
        g = sns.FacetGrid(sample_data, hue="pop", col="sampleID",
            col_wrap=3, height=3) 
        g.map(plt.scatter, "x", "y", s=50)
        g.add_legend()

        if save:
            plt.savefig(os.path.join(self.output_folder, "location_plot.png"))

        plt.close()


    def plot_contour_map(self, sampleID=None):
        """
        Plots contour map for a given sampleID. If no sampleID is provided,
        all samples of unknown origin will be plotted.

        Parameters
        ----------
        sampleID : str, optional
            The sampleID to plot. If None, all samples will be plotted.
            The default is None.
        
        Returns
        -------
        None.
        """
        if self.contour_classification is None:
            raise ValueError("Classification results not available. Please run " +
                "classify_by_contours() before plotting contour maps.")

        if sampleID is None:
            sample_list = self.data.unknowns.sampleID.unique()
        else:
            sample_list = [sampleID]

        image_list = []

        for sample in sample_list:    
            image = mpimg.imread(os.path.join(self.output_folder,
                                "contour_" + sample + ".png"))
            image_list.append(image)

        cols = min(3, len(sample_list))
        rows = int(np.ceil(len(image_list) / cols))
        fig, axs = plt.subplots(rows, cols)
        plt.subplots_adjust(wspace=0.00, hspace=0.00)

        if (rows == 1 & cols == 1):
            axs.spines['top'].set_visible(False)
            axs.spines['right'].set_visible(False)
            axs.spines['bottom'].set_visible(False)
            axs.spines['left'].set_visible(False)
            axs.get_xaxis().set_ticks([])
            axs.get_yaxis().set_ticks([])
            axs.imshow(image_list[0])
        else:
            for i, ax in enumerate(axs.flat):
                if i < len(image_list):
                    ax.spines['top'].set_visible(False)
                    ax.spines['right'].set_visible(False)
                    ax.spines['bottom'].set_visible(False)
                    ax.spines['left'].set_visible(False)
                    ax.get_xaxis().set_ticks([])
                    ax.get_yaxis().set_ticks([])
                    ax.imshow(image_list[i])
                else:
                    ax.axis("off")

        plt.show()
        plt.close()
        

    def plot_confusion_matrix(self, save=True):
        """
        Plots confusion matrix. This functions uses the true and predicted
        labels generated from the test data to give a visual representation
        of the accuracy of the model.

        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        
        Returns
        -------
        None.
        """
        if self.classification_test_results is None:
            raise ValueError("No classification results to plot.")

        _plot_confusion_matrix(self.classification_test_results,
            self.classification_confusion_matrix,
            self.nn_type, self.output_folder, save)


    def plot_assignment(self, save=True, col_scheme="Spectral"):
        """
        Plots the proportion of times each individual from the
        unknown data was assigned to each population.

        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        col_scheme : str, optional
            The colour scheme to use for the plot. The default is "Spectral".

        Returns
        -------
        None
        """
        if self.contour_classification is None:
            raise ValueError("No classification results to plot.")

        e_preds = self.contour_classification.copy()

        _plot_assignment(e_preds, col_scheme, self.output_folder,
            self.nn_type, save)


    def plot_structure(self, save=True, col_scheme="Spectral"):
        """
        Plots the proportion of times individuals from the
        test data were assigned to the correct population. 
        Used for determining the accuracy of the classifier.

        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        col_scheme : str, optional
            The colour scheme to use for the plot. The default is "Spectral".
        
        Returns
        -------
        None
        """
        classes = np.unique(self.classification_test_results["true_pop"])
        preds = pd.DataFrame(self.classification_confusion_matrix,
                             columns=classes,
                             index=classes)

        _plot_structure(preds, col_scheme, self.nn_type, 
            self.output_folder, save)

    def save(self, save_path=None, filename="regressor.pkl"):
        """
        Saves the current instance of the class to a pickle file.

        Parameters
        ----------
        save_path : str, optional
            The path to save the file to. If None, the file will be saved
            to the current working directory. The default is None.
        filename : str, optional
            The name of the file to save. The default is "regressor.pkl".

        Returns
        -------
        None.
        """
        _save(self, save_path, filename)

    @staticmethod
    def load(load_path=None):
        """
        Loads a saved instance of the class from a pickle file.

        Parameters
        ----------
        load_path : str, optional
            The path to load the file from. The default is None.
        
        Returns
        -------
        None
        """
        return _load(load_path)

    # Hidden functions below
    def _fit_regressor_model(self, epochs, train_loader, valid_loader, 
                             net, optimizer, scheduler, loss_func):

        loss_dict = {"epoch": [], "train": [], "valid": []}

        # Testing
        # model = nn.Sequential(
        #     nn.Linear(2, 10),
        #     nn.ReLU(),
        #     nn.Linear(10, 1)
        # )
        ####

        for epoch in range(epochs):

            train_loss = 0
            valid_loss = 0

            for _, (data, target) in enumerate(train_loader):
                optimizer.zero_grad()
                output = net(data)
                loss = loss_func(output.squeeze(), target.squeeze().long())
                loss.backward()
                optimizer.step()
                train_loss += loss.data.item()
        
            # Calculate average train loss
            avg_train_loss = train_loss / len(train_loader)

            for _, (data, target) in enumerate(valid_loader):
                output = net(data)
                loss = loss_func(output.squeeze(), target.squeeze().long())
                valid_loss += loss.data.item()

                if valid_loss < self.lowest_val_loss:
                    self.__lowest_val_loss = valid_loss
                    torch.save(net, os.path.join(self.output_folder, "best_model.pt"))

            # Step scheduler for LR decay
            scheduler.step(valid_loss)

            # Calculate average validation loss
            avg_valid_loss = valid_loss / len(valid_loader)

            loss_dict["epoch"].append(epoch)
            loss_dict["train"].append(avg_train_loss)
            loss_dict["valid"].append(avg_valid_loss)

        return pd.DataFrame(loss_dict)


    def _euclidean_dist_loss(self, y_pred, y_true):

        y_pred = y_pred.detach().numpy()
        y_true = y_true.detach().numpy()

        loss = np.sqrt(np.sum(np.square(y_pred - y_true)))
        loss = Variable(torch.tensor(loss), requires_grad=True)

        return loss

    def _unnormalize_locations(self, y_pred):

        y_pred_unnorm = np.array(
            [[x[0] * self.data.sdlong + self.data.meanlong,
              x[1] * self.data.sdlat + self.data.meanlat
            ] for x in y_pred])

        return y_pred_unnorm
    
    def _normalize_locations(self, y_pred):

        y_pred_norm = np.array(
            [[(x[0] - self.data.meanlong) / self.data.sdlong,
              (x[1] - self.data.meanlat) / self.data.sdlat
            ] for x in y_pred])

        return y_pred_norm

    def _generate_bootstrap_results(self, nboots, nreps, epochs, valid_size,
                             cv_splits, cv_reps, learning_rate, batch_size,
                             dropout_prop, jobs):

        # Create tempfolder
        tempfolder = tempfile.mkdtemp()
        self.save(save_path=tempfolder)

        # Find path to _multiboots
        filepath = pf.__file__
        folderpath = os.path.dirname(filepath)

        # run multiple bootstraps in parallel using the mb script
        call(["python", folderpath + "/_multiboots.py", "-p", tempfolder,
              "-n", str(nboots), "-r", str(nreps), "-e", str(epochs),
              "-v", str(valid_size), "-s", str(cv_splits), "-c", str(cv_reps),
              "-l", str(learning_rate), "-b", str(batch_size), "-d", str(dropout_prop),
              "-j", str(jobs)])

        self.test_locs_final = pd.read_csv(os.path.join(tempfolder, "test_locs_final.csv"))
        self.pred_locs_final = pd.read_csv(os.path.join(tempfolder, "pred_locs_final.csv"))      


    def _test_classification(self, test_locs, num_contours, save_plots):

        test_report = {"sampleID": [], "true_pop": [],
                       "pred_pop": [], "kd_estimate": []}

        for sample in test_locs["sampleID"].unique():

            sample_df = test_locs[test_locs["sampleID"] == sample]

            X, Y, Z, xlim, ylim = self._generate_kde(sample_df)
            cset = self._find_cset_from_contours(X, Y, Z, xlim,
                ylim, test_locs, num_contours, sample, save_plots)

            # Find predicted pop
            pred_pop, kd = self._contour_finder(test_locs, cset)
            test_report["sampleID"].append(sample)
            test_report["true_pop"].append(np.unique(sample_df["pop"])[0])
            test_report["pred_pop"].append(pred_pop)
            test_report["kd_estimate"].append(kd)

        return pd.DataFrame(test_report)

    def _classify_unknowns(self, pred_locs, test_locs, num_contours, save_plots):

        classification_data = {"sampleID": [], "classification": [], "kd_estimate": []}

        for sample in pred_locs["sampleID"].unique():

            sample_df = pred_locs[pred_locs["sampleID"] == sample]

            X, Y, Z, xlim, ylim = self._generate_kde(sample_df)
            cset = self._find_cset_from_contours(X, Y, Z, xlim,
                ylim, test_locs, num_contours, sample, save_plots)

            # Find predicted pop
            pred_pop, kd = self._contour_finder(test_locs, cset)
            classification_data["sampleID"].append(sample)
            classification_data["classification"].append(pred_pop)
            classification_data["kd_estimate"].append(kd)

        return pd.DataFrame(classification_data)

    def _contour_finder(self, true_dat, cset):
        """
        Finds population in densest contour.

        Parameters
        ----------
        true_dat : pd.DataFrame
            Dataframe containing x and y coordinates of all populations in
            training set.
        cset : matplotlib.contour.QuadContourSet
            Contour values for each contour polygon.

        Returns
        pred_pop : string
            Name of population in densest contour.
        """

        cont_dict = {"pop": [], "cont": []}

        for pop in true_dat["pop"].values:
            cont_dict["pop"].append(pop)
            cont = 0
            point = np.array(
                [
                    [
                        true_dat[true_dat["pop"] == pop]["x"].values[0],
                        true_dat[true_dat["pop"] == pop]["y"].values[0],
                    ]
                ]
            )

            for i in range(1, len(cset.allsegs)):
                for j in range(len(cset.allsegs[i])):
                    path = matplotlib.path.Path(cset.allsegs[i][j].tolist())
                    inside = path.contains_points(point)
                    if inside[0]:
                        cont = i
                        break
                    else:
                        next
            cont_dict["cont"].append(np.round(cset.levels[cont], 2))

        pred_pop = cont_dict["pop"][np.argmin(cont_dict["cont"])]

        return pred_pop, min(cont_dict["cont"])

    def _generate_kde(self, sample_df):

        d_x = (max(sample_df["x_pred"]) - min(sample_df["x_pred"])) / 5
        d_y = (max(sample_df["y_pred"]) - min(sample_df["y_pred"])) / 5
        xlim = min(sample_df["x_pred"]) - d_x, max(sample_df["x_pred"]) + d_x
        ylim = min(sample_df["y_pred"]) - d_y, max(sample_df["y_pred"]) + d_y

        X, Y = np.mgrid[xlim[0]:xlim[1]:100j, ylim[0]:ylim[1]:100j]

        positions = np.vstack([X.ravel(), Y.ravel()])
        values = np.vstack([sample_df["x_pred"], sample_df["y_pred"]])

        try:
            kernel = stats.gaussian_kde(values)
        except (ValueError) as e:
            raise Exception("Too few points to generate contours") from e

        Z = np.reshape(kernel(positions).T, X.shape)
        new_z = Z / np.max(Z)

        return X, Y, new_z, xlim, ylim

    def _find_cset_from_contours(self, X, Y, Z, xlim, ylim, test_locs, num_contours,
                       sample, save):

        # Plot
        plt.ioff()
        fig = plt.figure(figsize=(8, 8))
        ax = fig.gca()
        plt.xlim(xlim[0], xlim[1])
        plt.ylim(ylim[0], ylim[1])
        cset = ax.contour(X, Y, Z, levels=num_contours, colors="black")

        cset.levels = -np.sort(-cset.levels)

        for pop in np.unique(test_locs["pop"].values):
            x = test_locs[test_locs["pop"] == pop]["x"].values[0]
            y = test_locs[test_locs["pop"] == pop]["y"].values[0]
            plt.scatter(x, y, cmap=plt.cm.Spectral, label=pop)

        ax.clabel(cset, cset.levels, inline=1, fontsize=10)
        ax.set_xlabel("Longitude")
        ax.set_ylabel("Latitude")
        plt.title(sample)
        plt.legend()

        if save:
            plt.savefig(self.output_folder + "/contour_" + \
                        sample + ".png", format="png")

        plt.close()

        return cset

    # Validation functions
    def _validate_init_inputs(self, data, nboots, random_state, output_folder):
        if not isinstance(data, GeneticData):
            raise TypeError("data must be an instance of GeneticData")

        if not isinstance(nboots, int):
            raise TypeError("nboots must be an integer")

        if not isinstance(random_state, int):
            raise TypeError("random_state must be an integer")

        if output_folder is not None:
            if not isinstance(output_folder, str):
                raise TypeError("output_folder must be a string")

            if not os.path.isdir(output_folder):
                raise ValueError("output_folder must be a valid directory")

    def _validate_train_inputs(self, epochs, valid_size, cv_splits, cv_reps,
                            learning_rate, batch_size, dropout_prop):

        if not isinstance(epochs, int):
            raise TypeError("epochs must be an integer")
        
        if not isinstance(valid_size, float):
            raise TypeError("valid_size must be a float")

        if valid_size > 1 or valid_size < 0:
            raise ValueError("valid_size must be between 0 and 1")
        
        if not isinstance(cv_splits, int):
            raise TypeError("cv_splits must be an integer")

        if not isinstance(cv_reps, int):
            raise TypeError("cv_reps must be an integer")

        if not isinstance(learning_rate, float):
            raise TypeError("learning_rate must be a float")

        if learning_rate > 1 or learning_rate < 0:
            raise ValueError("learning_rate must be between 0 and 1")

        if not isinstance(batch_size, int):
            raise TypeError("batch_size must be an integer")

        if not isinstance(dropout_prop, float) and not isinstance(dropout_prop, int):
            raise TypeError("dropout_prop must be a float")

        if dropout_prop > 1 or dropout_prop < 0:
            raise ValueError("dropout_prop must be between 0 and 1")

    def _validate_contour_inputs(self, nboots, num_contours, save_plots, save):
            
        if not isinstance(nboots, int):
            raise TypeError("nboots must be an integer")

        if not isinstance(num_contours, int):
            raise TypeError("num_contours must be an integer")

        if not isinstance(save_plots, bool):
            raise TypeError("save_plots must be a boolean")

        if not isinstance(save, bool):
            raise TypeError("save must be a boolean")


import numpy as np
import pandas as pd
import allel # change to sgkit eventually
import zarr
import h5py
import sys
import os

from sklearn.model_selection import RepeatedStratifiedKFold, train_test_split
from sklearn.preprocessing import LabelEncoder

class GeneticData():
    """
    Class for reading in genetic data and sample data and compiling
    information into a pandas DataFrame for either a classifier or
    regressor neural network.

    Parameters
    ----------
    genetic_data : str
        Path to genetic data file. Can be .zarr, .vcf, or .h5py.
    sample_data : str
        Path to sample data file. Must be .tsv or .txt.
    test_size : float
        Proportion of data to be used for testing.
    seed : int
        Seed for random number generator.
    
    Attributes
    ----------
    genetic_data : str
        Path to genetic data file. Can be .zarr, .vcf, or .h5py.
    sample_data : str
        Path to sample data file. Must be .tsv or .txt and contain the 
        columns "x" (longitude), "y" (latitude), "pop" (population name),
        and "sampleID". The "sampleID" must match the sample IDs in 
        the genetic data file. 
    seed : int
        Seed for random number generator.
    meanlong : float
        Mean longitude of sample data.
    meanlat : float
        Mean latitude of sample data.
    stdlon : float
        Standard deviation of longitude of sample data.
    stdlat : float
        Standard deviation of latitude of sample data.
    
    Methods
    -------
    read_data()
        Reads a .zarr, .vcf, or h5py file containing genetic data and
        compiles information into a pandas DataFrame for either a
        classifier or regressor neural network.
    split_train_test(data, test_size, seed)
        Splits data into training and testing sets.
    split_kfcv(data, n_splits, n_reps, seed)
        Splits data into k-fold cross validation sets.
    split_unknowns(data)
        Splits data into known and unknown samples.
    """
    def __init__(self, genetic_data=None, sample_data=None, test_size=0.2, seed=123):

        self._validate_init_inputs(genetic_data, sample_data, test_size, seed)

        self.genetic_data = genetic_data
        self.sample_data = sample_data
        self.seed = seed
        self.meanlong = None
        self.meanlat = None
        self.stdlon = None
        self.stdlat = None

        if genetic_data is not None and sample_data is not None:
            self._initialize(test_size=test_size, seed=seed)

    def read_data(self):
        """
        Reads a .zarr, .vcf, or h5py file containing genetic data and
        compiles information into a pandas DataFrame for either a 
        classifier or regressor neural network.

        Returns
        -------
        data : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information.
        """

        self._validate_read_data_inputs()

        # Load genotypes
        print("loading genotypes")
        samples, dc = self._load_genotypes(self.genetic_data)

        # Load data and organize for output
        print("loading sample data")
        locs = pd.read_csv(self.sample_data, sep="\t")
        locs = self._sort_samples(locs, samples)
        locs["alleles"] = list(dc)

        # Normalize location data
        self._retrieve_summary_stats(locs)
        locs = self._normalize_locations(locs)

        return locs

    def split_unknowns(self, data):
        """
        Splits data into known and unknown samples.
        
        Parameters
        ----------
        data : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information. This is the output of `read_data()`.
        
        Returns
        -------
        known : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information for known samples.
        unknown : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information for unknown samples.
        """
        # Split data into known and unknown
        unknown = data[data["pop"].isnull()]
        known = data[data["pop"].notnull()]

        return known, unknown

    def split_train_test(self, data=None, stratify_by_pop=True, test_size=0.2, seed=123, bootstrap=False):
        """
        Splits data into training and testing sets.
        
        Parameters
        ----------
        data : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information. This is the output of `read_data()`.  
        stratify_by_pop : bool
            Whether to stratify the data by population. Default is True.
        test_size : float
            Proportion of data to be used for testing. Default is 0.2.
        seed : int
            Random seed for reproducibility. Default is 123.
        bootstrap : bool
            Whether to bootstrap the training data. Default is False.
            
        Returns
        -------
        train : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information for training samples.
        test : Pandas DataFrame
            Contains information on corresponding sampleID and
            genetic information for testing samples.
        """
        # Split data into training and testing
        if stratify_by_pop is True:
            train, test = self._stratified_split(data, test_size=test_size, seed=seed)
        else:
            train, test = self._random_split(data, test_size=test_size, seed=seed)

        if bootstrap:
            train = train.sample(frac=1, replace=True, random_state=seed)

        return train, test

    def split_kfcv(self, data=None, n_splits=5, n_reps=1, seed=123, stratify_by_pop=True, bootstrap=False):
        """
        Splits data into training and testing sets.

        Parameters
        ----------
        n_splits : int
            Number of splits for cross-validation. Default is 5.
        n_reps : int
            Number of repetitions for cross-validation. Default is 1.
        stratify_by_pop : bool
            Whether to stratify the data by population. Default is True.
        seed : int
            Seed for random number generator. Default is 123.
        bootstrap : bool
            Whether to bootstrap the training data. Default is False.

        Returns
        -------
        list of tuples
            Each tuple contains a training and testing set. The length
            of the list is n_splits * n_reps.
        """
        rskf = RepeatedStratifiedKFold(n_splits=n_splits, n_repeats=n_reps,
                                       random_state=seed)
        dataset_list = []

        if data is None:
            known_data = self.knowns.copy()
        else:
            known_data = data

        if stratify_by_pop:
            for _, (train_ind, test_ind) in enumerate(rskf.split(known_data["alleles"],
                                                                 known_data["pop"], 
                                                                 groups=known_data["pop"])):
                train = known_data.iloc[train_ind]
                test = known_data.iloc[test_ind]

                if bootstrap:
                    train = train.sample(frac=1, replace=True, random_state=seed)

                dataset_tuple = (train, test)
                dataset_list.append(dataset_tuple)
        else:
            for _, (train_ind, test_ind) in enumerate(rskf.split(known_data["alleles"], 
                                                                 known_data["pop"])):
                train = known_data.iloc[train_ind]
                test = known_data.iloc[test_ind]

                if bootstrap:
                    train = train.sample(frac=1, replace=True, random_state=seed)
                    
                dataset_tuple = (train, test)
                dataset_list.append(dataset_tuple)

        return dataset_list

    def update_unknowns(self, new_genetic_data, new_sample_data):
        """
        Reads a .zarr, .vcf, or h5py file containing genetic data 
        of new samples from unknown origins, replacing the old
        data for samples from unknown origins stored in the GeneticData
        object.

        Parameters
        ----------
        new_genetic_data : str
            Path to genetic data file. Can be .zarr, .vcf, or .h5py.
        new_sample_data : str
            Path to sample data file. Must be .tsv or .txt and contain the 
            columns "x" (longitude), "y" (latitude), "pop" (population name),
            and "sampleID". The "sampleID" must match the sample IDs in 
            the genetic data file. The "pop" column should be filled with 
            NAs since the new samples are of unknown origin.

        Returns
        -------
        None
        """

        self._validate_update_unknowns_inputs(new_genetic_data, new_sample_data)

        # Load genotypes
        print("loading genotypes")
        samples, dc = self._load_genotypes(new_genetic_data)

        # Load data and organize for output
        print("loading sample data")
        locs = pd.read_csv(new_sample_data, sep="\t")
        locs = self._sort_samples(locs, samples)
        locs["alleles"] = list(dc)

        # Normalize location data
        self._retrieve_summary_stats(locs)
        new_data = self._normalize_locations(locs)

        # Reset unknowns and full data
        _, self.unknowns = self.split_unknowns(new_data)
        self.data = pd.concat([self.knowns, self.unknowns], ignore_index=True)

    def _initialize(self, test_size=0.2, seed=123):

        self.data = self.read_data()
        self.knowns, self.unknowns = self.split_unknowns(self.data)
        self.train, self.test = self.split_train_test(test_size=test_size, seed=seed)

        # Create label encoder from train target
        self.label_enc = LabelEncoder()
        self.label_enc.fit_transform(self.train["pop"])

    def _load_genotypes(self, genetic_data):

        if genetic_data.endswith(".zarr"):
            callset = zarr.open_group(genetic_data, mode="r")
            gt = callset["calldata/GT"]
            gen = allel.GenotypeArray(gt[:])
            samples = callset["samples"][:]

        elif genetic_data.endswith(".vcf") or genetic_data.endswith(".vcf.gz"):
            vcf = allel.read_vcf(genetic_data, log=sys.stderr)
            gen = allel.GenotypeArray(vcf["calldata/GT"])
            samples = vcf["samples"]

        elif genetic_data.endswith(".hdf5"):
            h5 = h5py.File(genetic_data, "r")
            dc = np.array(h5["derived_counts"])
            samples = np.array(h5["samples"])
            h5.close()

        else:
            raise ValueError("genetic_data must have extension 'zarr', 'vcf', or 'hdf5'")

        # count derived alleles for biallelic sites
        if genetic_data.endswith(".hdf5") is False:

            print("counting alleles")
            ac = gen.to_allele_counts()
            biallel = gen.count_alleles().is_biallelic()
            dc = np.array(ac[biallel, :, 1], dtype="int_")
            dc = np.transpose(dc)

        return samples, dc

    def _sort_samples(self, locs, samples):

        if not pd.Series([
            "x", "pop", "y", "sampleID"
            ]).isin(locs.columns).all():
            raise ValueError("sample_data does not have correct columns")

        locs["id"] = locs["sampleID"]
        locs.set_index("id", inplace=True)

        # sort loc table so samples are in same order as genotype samples
        locs = locs.reindex(np.array(samples))

        # Create order column for indexing
        locs["order"] = np.arange(0, len(locs))

        # check that all sample names are present
        if not all(
            [locs["sampleID"][x] == samples[x] for x in range(len(samples))]
        ):
            raise ValueError(
                "sample ordering failed! Check that sample IDs match VCF.")

        return locs

    def _stratified_split(self, data, test_size=0.2, seed=123):

        if data is None:
            data = self.knowns

        X_train, X_test, y_train, y_test = train_test_split(
            data["alleles"], data[["x", "y", "pop"]],
            stratify=data["pop"],
            random_state=seed, test_size=test_size)

        train = pd.concat([X_train, y_train], axis=1)
        test = pd.concat([X_test, y_test], axis=1)

        return train, test

    def _random_split(self, test_size=0.2, seed=123):

        if data is None:
            data = self.knowns

        X_train, X_test, y_train, y_test = train_test_split(
        data["alleles"], data[["x", "y", "pop"]],
        random_state=seed, test_size=test_size)

        train = pd.concat([X_train, y_train], axis=1)
        test = pd.concat([X_test, y_test], axis=1)

        return train, test

    def _retrieve_summary_stats(self, raw_data):

        self.meanlong = np.nanmean(raw_data['x'])
        self.sdlong = np.nanstd(raw_data['x'])
        self.meanlat = np.nanmean(raw_data['y'])
        self.sdlat = np.nanstd(raw_data['y'])

    def _normalize_locations(self, raw_data):
        """
        Normalize location corrdinates.

        Parameters
        ----------
        raw_data : pd.DataFrame
            A pandas DataFrame corresponding to the results from `read_data()`.

        Returns
        -------
        data : pd.DataFrame
            A pandas DataFrame with normalized location coordinates.
        """
        raw_data["x_norm"] = (raw_data['x'].tolist() - self.meanlong) / self.sdlong
        raw_data["y_norm"] = (raw_data['y'].tolist() - self.meanlat) / self.sdlat

        return raw_data

    def _unnormalize_locations(self, data):
        """
        Unnormalize location corrdinates.

        Parameters
        ----------
        data : pd.DataFrame
            A pandas DataFrame corresponding to the results from `read_data()`.

        Returns
        -------
        data : pd.DataFrame
            A pandas DataFrame with unnormalized location coordinates.
        """
        data["x"] = data['x_norm'].tolist() * self.sdlong + self.meanlong
        data["y"] = data['y_norm'].tolist() * self.sdlat + self.meanlat

        return data

    def _validate_init_inputs(self, genetic_data, sample_data, test_size, seed):

        if genetic_data is not None and not isinstance(genetic_data, str):
            raise ValueError("genetic_data must be a string")

        if sample_data is not None and not isinstance(sample_data, str):
            raise ValueError("sample_data must be a string")

        if not isinstance(test_size, float):
            raise ValueError("test_size must be a float")

        if test_size > 1 or test_size < 0:
            raise ValueError("test_size must be between 0 and 1")

        if not isinstance(seed, int):
            raise ValueError("seed must be an integer")

    def _validate_read_data_inputs(self):

        if self.genetic_data is None:
            raise ValueError("genetic_data is None")

        if self.sample_data is None:
            raise ValueError("sample_data is None")

        if os.path.exists(self.genetic_data) is False:
            raise ValueError("Path to genetic_data does not exist")

        if os.path.exists(self.sample_data) is False:
            raise ValueError("Path to sample_data does not exist")

        if self.genetic_data.endswith((".zarr", ".vcf", ".hdf5")) is False:
            raise ValueError("genetic_data must have extension 'zarr', 'vcf', or 'hdf5'")

        if self.sample_data.endswith((".txt", ".tsv")) is False:
            raise ValueError("sample_data must have extension 'txt' or 'tsv'")

        locs = pd.read_csv(self.sample_data, sep="\t")
        locs_list = locs.columns.tolist()
        if set(locs_list) != set(["x", "pop", "y", "sampleID"]):
            raise ValueError("sample_data file does not have correct columns")

    def _validate_update_unknowns_inputs(self, new_genetic_data, new_sample_data):

        if not isinstance(new_genetic_data, str):
            raise ValueError("new_unknowns must be a path to a genetic data file")

        if not isinstance(new_sample_data, str):
            raise ValueError("new_unknowns must be a path to a sample data file")

        if os.path.exists(new_genetic_data) is False:
            raise ValueError("Path to new_genetic_data does not exist")

        if os.path.exists(new_sample_data) is False:
            raise ValueError("Path to new_sample_data does not exist")

        if new_genetic_data.endswith((".zarr", ".vcf", ".hdf5")) is False:
            raise ValueError("new_genetic_data must have extension 'zarr', 'vcf', or 'hdf5'")

        if new_sample_data.endswith((".txt", ".tsv")) is False:
            raise ValueError("new_sample_data must have extension 'txt' or 'tsv'")

        locs = pd.read_csv(new_sample_data, sep="\t")
        locs_list = locs.columns.tolist()
        if set(locs_list) != set(["x", "pop", "y", "sampleID"]):
            raise ValueError("sample_data file does not have correct columns")

        




# Copyright (c) 2023 Apex Resource Management Solution Ltd. (ApexRMS). All rights reserved.
# MIT License

from setuptools import setup, find_packages
from pathlib import Path
from popfinder._version import __version__

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

entry_points = \
{'console_scripts': ['pop_classifier = popfinder.cli_classifier:main',
                     'pop_regressor = popfinder.cli_regressor:main']}

setup(name="popfinder",
      version=__version__,
      description="Genetic population assignment using neural networks",
      long_description=long_description,
      long_description_content_type='text/markdown',
      author="Katie Birchard",
      author_email="katie.birchard@apexrms.com",
      url=None,
      packages=find_packages(exclude="tests"),
      # install_requires=install_requires,
      entry_points=entry_points)

install_requires = ["numpy", "pandas", "torch", "scikit-learn", "dill",
                    "seaborn", "matplotlib", "scikit-allel", "zarr",
                    "h5py", "scipy"]


import pandas as pd
import numpy as np
import multiprocessing as mp
import argparse
import os

from popfinder.dataloader import GeneticData
from popfinder.classifier import PopClassifier
# from popfinder.regressor import PopRegressor

def _train_on_bootstraps(arg_list):

    popfinder_path, nboots, epochs, valid_size, cv_splits, cv_reps, learning_rate, batch_size, dropout_prop, rep = arg_list
    test_results_final = pd.DataFrame({"bootstrap": [], "sampleID": [], "pop": [], "x": [],
                                    "y": [], "x_pred": [], "y_pred": []})
    pred_results_final = pd.DataFrame({"bootstrap": [], "sampleID": [], "pop": [], 
                                    "x_pred": [], "y_pred": []})   
    for boot in range(nboots):

        popfinder = PopClassifier.load(os.path.join(popfinder_path, "classifier.pkl"))
        popfinder.output_folder = os.path.join(popfinder_path, "rep{}".format(rep))
        os.makedirs(popfinder.output_folder, exist_ok=True)

        # Use bootstrap to randomly select sites from training/test/unknown data
        num_sites = popfinder.data.train["alleles"].values[0].shape[0]

        site_indices = np.random.choice(range(num_sites), size=num_sites,
                                        replace=True)

        popfinder.__boot_data = GeneticData()
        popfinder.__boot_data.train = popfinder.data.train.copy()
        popfinder.__boot_data.test = popfinder.data.test.copy()
        popfinder.__boot_data.knowns = pd.concat([popfinder.data.train, popfinder.data.test])
        popfinder.__boot_data.unknowns = popfinder.data.unknowns.copy()

        # Slice datasets by site_indices
        popfinder.__boot_data.train["alleles"] = [a[site_indices] for a in popfinder.data.train["alleles"].values]
        popfinder.__boot_data.test["alleles"] = [a[site_indices] for a in popfinder.data.test["alleles"].values]
        popfinder.__boot_data.unknowns["alleles"] = [a[site_indices] for a in popfinder.data.unknowns["alleles"].values]

        # Train on new training set
        popfinder.data
        popfinder.train(epochs=epochs, valid_size=valid_size,
                cv_splits=cv_splits, cv_reps=cv_reps,
                learning_rate=learning_rate, batch_size=batch_size,
                dropout_prop=dropout_prop)
        popfinder.test()
        test_locs = popfinder.test_results.copy()
        # test_locs["sampleID"] = test_locs.index
        test_locs["bootstrap"] = boot

        if popfinder.data.unknowns.shape[0] > 0:

            pred_locs = popfinder.assign_unknown(save=False)
            pred_locs["bootstrap"] = boot
            pred_results_final = pd.concat([pred_results_final,
                pred_locs[["bootstrap", "sampleID", "true_pop", "pred_pop"]]])

        # test_results_final = pd.concat([test_results_final,
        #     test_locs[["bootstrap", "sampleID", "pop", "x", "y", "x_pred", "y_pred"]]])
        test_results_final = pd.concat([test_results_final,
                                        test_locs[["bootstrap", "true_pop", "pred_pop"]]])
        
    return test_results_final, pred_results_final


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", help="Path to PopClassifier object")
    parser.add_argument("-n", help="Number of bootstraps", type=int)
    parser.add_argument("-r", help="Number of repetitions", type=int)
    parser.add_argument("-e", help="Number of epochs", type=int)
    parser.add_argument("-v", help="Validation size", type=float)
    parser.add_argument("-s", help="Number of cross-validation splits", type=int)
    parser.add_argument("-c", help="Number of cross-validation repetitions", type=int)
    parser.add_argument("-l", help="Learning rate", type=float)
    parser.add_argument("-b", help="Batch size", type=int)
    parser.add_argument("-d", help="Dropout proportion", type=float)
    parser.add_argument("-j", help="Number of jobs", type=int)
    args = parser.parse_args()
    popfinder_path = args.p
    nboots = args.n
    nreps = args.r
    epochs = args.e 
    valid_size = args.v 
    cv_splits = args.s
    cv_reps = args.c 
    learning_rate = args.l 
    batch_size = args.b
    dropout_prop = args.d
    num_jobs = args.j

    if num_jobs == -1:
        num_jobs = mp.cpu_count()
    pool = mp.Pool(processes=num_jobs)
    results = pool.map(_train_on_bootstraps, [[popfinder_path, nboots, epochs, valid_size, cv_splits,
                                               cv_reps, learning_rate, batch_size,
                                               dropout_prop, rep] for rep in range(nreps)])
    pool.close()
    pool.join()

    test_locs_final = pd.DataFrame()
    pred_locs_final = pd.DataFrame()
    for rep in range(nreps):
        test_locs = results[rep][0]
        pred_locs = results[rep][1]
        test_locs["rep"] = rep
        pred_locs["rep"] = rep
        test_locs_final = pd.concat([test_locs_final, test_locs])
        pred_locs_final = pd.concat([pred_locs_final, pred_locs])

    test_locs_final.to_csv(os.path.join(popfinder_path, "test_locs_final.csv"), index=False)
    pred_locs_final.to_csv(os.path.join(popfinder_path, "pred_locs_final.csv"), index=False)


import torch
import torch.nn as nn
from sklearn.metrics import accuracy_score, confusion_matrix, f1_score, precision_score, recall_score
from sklearn.preprocessing import OneHotEncoder
from collections import Counter
import numpy as np
import pandas as pd
import os
import shutil
from subprocess import call
import tempfile

import popfinder as pf
from popfinder.dataloader import GeneticData
from popfinder._neural_networks import ClassifierNet
from popfinder._helper import _generate_train_inputs
from popfinder._helper import _generate_data_loaders
from popfinder._helper import _data_converter
from popfinder._helper import _split_input_classifier
from popfinder._helper import _save, _load
from popfinder._visualize import _plot_assignment
from popfinder._visualize import _plot_training_curve
from popfinder._visualize import _plot_confusion_matrix
from popfinder._visualize import _plot_structure

pd.options.mode.chained_assignment = None

class PopClassifier(object):
    """
    A class to represent a classifier neural network object for population assignment.
    """
    def __init__(self, data, random_state=123, output_folder=None):

        self._validate_init_inputs(data, random_state, output_folder)

        self.__data = data # GeneticData object
        self.__random_state = random_state
        if output_folder is None:
            output_folder = os.getcwd()
        self.__output_folder = output_folder
        self.__cv_output_folder = os.path.join(output_folder, "cv_results")
        self.__label_enc = data.label_enc
        self.__train_history = None
        self.__best_model = None
        self.__test_results = None # use for cm and structure plot
        self.__classification = None # use for assignment plot
        self.__accuracy = None
        self.__precision = None
        self.__recall = None
        self.__f1 = None
        self.__confusion_matrix = None
        self.__nn_type = "classifier"
        self.__mp_run = False
        self.__lowest_val_loss_total = 9999

    @property
    def data(self):
        return self.__data

    @property
    def random_state(self):
        return self.__random_state
    
    @property
    def output_folder(self):
        return self.__output_folder

    @output_folder.setter
    def output_folder(self, output_folder):
        self.__output_folder = output_folder
        self.__cv_output_folder = os.path.join(output_folder, "cv_results")

    @property
    def label_enc(self):
        return self.__label_enc

    @label_enc.setter
    def label_enc(self, value):
        self.__label_enc = value

    @property
    def train_history(self):
        return self.__train_history

    @property
    def best_model(self):
        return self.__best_model

    @property
    def test_results(self):
        return self.__test_results
    
    @property
    def cv_test_results(self):
        return self.__cv_test_results

    @property
    def classification(self):
        return self.__classification

    @property
    def accuracy(self):
        return self.__accuracy

    @property
    def precision(self):
        return self.__precision

    @property
    def recall(self):
        return self.__recall

    @property
    def f1(self):
        return self.__f1
    
    @property
    def confusion_matrix(self):
        return self.__confusion_matrix

    @property
    def cv_accuracy(self):
        return self.__cv_accuracy

    @property
    def cv_precision(self):
        return self.__cv_precision

    @property
    def cv_recall(self):
        return self.__cv_recall

    @property
    def cv_f1(self):
        return self.__cv_f1
    
    @property
    def cv_confusion_matrix(self):
        return self.__cv_confusion_matrix

    @property
    def nn_type(self):
        return self.__nn_type

    def train(self, epochs=100, valid_size=0.2, cv_splits=1, nreps=1,
              learning_rate=0.001, batch_size=16, dropout_prop=0, bootstraps=None,
              jobs=1, overwrite_results=True):
        """
        Trains the classification neural network.

        Parameters
        ----------
        epochs : int, optional
            Number of epochs to train the neural network. The default is 100.
        valid_size : float, optional
            Proportion of data to use for validation. The default is 0.2.
        cv_splits : int, optional
            Number of cross-validation splits. The default is 1.
        nreps : int, optional
            Number of repetitions. The default is 1.
        learning_rate : float, optional
            Learning rate for the neural network. The default is 0.001.
        batch_size : int, optional
            Batch size for the neural network. The default is 16.
        dropout_prop : float, optional
            Dropout proportion for the neural network. The default is 0.
        bootstraps : int, optional
            Number of bootstraps to perform. The default is None.
        jobs : int, optional
            If greater than 1, will use multiprocessing to train the neural network. 
            The default is 1.
        overwrite_results : boolean, optional
            If True, then will clear the output folder before training the new 
            model. The default is True.
        
        Returns
        -------
        None.
        """
        self._validate_train_inputs(epochs, valid_size, cv_splits, nreps,
                                    learning_rate, batch_size, dropout_prop)
        
        self.__prepare_result_folder(self.output_folder, overwrite_results)

        files = os.listdir(self.output_folder)
        if (overwrite_results) or (len(files) == 0) or (self.train_history is None):
            nrep_begin = 0
            self.__lowest_val_loss_total = 9999 # reset lowest val loss
            self.__train_history = None # reset train history
        else:
            existing_reps = [int(f.split("_")[-2].replace("rep", "")) for f in files if "rep" in f]
            nrep_begin = max(existing_reps)
            nreps = nrep_begin + nreps 

        multi_output = (bootstraps is not None) or (nreps is not None)

        if multi_output:

            if bootstraps is None:
                bootstraps = 1
            if nreps is None:
                nreps = nrep_begin + 1

            loss_df = pd.DataFrame()

            if jobs == 1:
                for i in range(bootstraps):
                    for j in range(nrep_begin, nreps):
                        #TODO: how does this affect mp results
                        if not self.__mp_run:
                            boot_folder = os.path.join(self.output_folder, f"rep{j+1}_boot{i+1}")
                            if not os.path.exists(boot_folder):
                                os.makedirs(boot_folder)
                        else:
                            boot_folder = self.output_folder

                        inputs = _generate_train_inputs(self.data, valid_size, cv_splits,
                                        nreps, seed=self.random_state, bootstrap=True)
                        boot_loss_df = self.__train_on_inputs(inputs, cv_splits, epochs, learning_rate,
                                            batch_size, dropout_prop, result_folder = boot_folder, 
                                            overwrite_results=overwrite_results)
                        
                        boot_loss_df.to_csv(os.path.join(boot_folder, "loss.csv"), index=False)
                        boot_loss_df["rep"] = j + 1
                        boot_loss_df["bootstrap"] = i + 1
                        loss_df = pd.concat([loss_df, boot_loss_df], axis=0, ignore_index=True)
            elif jobs > 1:
                # Create tempfolder
                tempfolder = tempfile.mkdtemp()

                # Let popfinder know this is a multiprocessing run (affects output folder creation)
                self.__mp_run = True
                self.save(save_path=tempfolder)

                # Find path to _mp_training
                filepath = pf.__file__
                folderpath = os.path.dirname(filepath)

                # Instead of looping through bootstrap iteration, run in parallel
                # to speed up training
                call(["python", folderpath + "/_mp_training.py", "-p", tempfolder,
                    "-n", str(bootstraps), "--r_start", str(nrep_begin), "-r", str(nreps), 
                    "-e", str(epochs), "-v", str(valid_size), "-s", str(cv_splits), 
                    "-l", str(learning_rate), "-b", str(batch_size), "-d", str(dropout_prop),
                    "-j", str(jobs)])
                loss_df = pd.read_csv(os.path.join(tempfolder, "train_history.csv"))

        # TODO: this never gets called
        else:
            inputs = _generate_train_inputs(self.data, valid_size, cv_splits,
                                            nreps, seed=self.random_state, bootstrap=False)
            loss_df = self.__train_on_inputs(inputs, cv_splits, epochs, learning_rate,
                                                batch_size, dropout_prop, result_folder = self.__cv_output_folder)

        # Save training history
        if self.__train_history is None:
            self.__train_history = loss_df
        else:
            self.__train_history = pd.concat([self.__train_history, loss_df], ignore_index=True)
       
       # Determine best model
        if (jobs == 1) or (not multi_output):
            self.__best_model = torch.load(os.path.join(self.output_folder, "best_model.pt"))
        else:
            best_model_folder, min_split = self.__find_best_model_folder_from_mp()
            self.__best_model = torch.load(os.path.join(best_model_folder, f"best_model_split{min_split}.pt"))
            torch.save(self.__best_model, os.path.join(self.output_folder, "best_model.pt"))
            self.__clean_mp_folders(nrep_begin, nreps, bootstraps)

    # TODO: move below
    def __find_best_model_folder_from_mp(self):
        min_loss = self.train_history.iloc[self.train_history[["valid"]].idxmin()]
        min_rep = min_loss["rep"].values[0]
        min_boot = min_loss["bootstrap"].values[0]
        min_split = min_loss["split"].values[0]
        best_model_folder = os.path.join(self.output_folder, f"rep{min_rep}_boot{min_boot}")
        return best_model_folder, min_split
    
    def __clean_mp_folders(self, nrep_begin, nreps, bootstraps):
        for rep in range(nrep_begin, nreps):
            for boot in range(bootstraps):
                folder = os.path.join(self.output_folder, f"rep{rep+1}_boot{boot+1}")
                os.remove(os.path.join(folder, "best_model.pt"))



    def test(self, use_best_model=True, save=True):
        """
        Tests the classification neural network.

        Parameters
        ----------
        use_best_model : bool, optional
            Whether to test using the best model only. If set to False, then will use all
            models generated from all training repeats and cross-validation splits and
            provide an ensemble frequency of assignments. The default is True.
        save : bool, optional
            Whether to save the test results to the output folder. The default is True.
        
        Returns
        -------
        None.
        """
        # Find unique reps/splits from cross validation
        reps = self.train_history["rep"].unique()
        splits = self.train_history["split"].unique()

        if "bootstrap" in self.train_history.columns:
            bootstraps = self.train_history["bootstrap"].unique()
        else:
            bootstraps = None
        
        test_input = self.data.test

        X_test = test_input["alleles"]
        y_test = test_input["pop"]

        y_test = self.label_enc.transform(y_test)
        X_test, y_test = _data_converter(X_test, y_test)

        y_true = y_test.squeeze()
        y_true_pops = self.label_enc.inverse_transform(y_true)

        # If not using just the best model, then test using all models
        if not use_best_model:
            if bootstraps is None: 
                bootstraps = 1
            if reps is None:
                reps = 1

            # Tests on all reps, bootstraps, and cv splits
            self.__test_results = self.__test_on_multiple_models(reps, bootstraps, splits, X_test, y_true_pops)
            y_pred = self.label_enc.transform(self.__test_results["pred_pop"])
            y_true = self.label_enc.transform(self.__test_results["true_pop"])
            y_true_pops = self.label_enc.inverse_transform(y_true)

        elif use_best_model:
            # Predict using the best model and revert from label encoder
            y_pred = self.best_model(X_test).argmax(axis=1)
            y_pred_pops = self.label_enc.inverse_transform(y_pred)

            self.__test_results = pd.DataFrame({"true_pop": y_true_pops,
                                                "pred_pop": y_pred_pops})

        if save:
            self.test_results.to_csv(os.path.join(self.output_folder,
                                    "classifier_test_results.csv"), index=False)

        self.__calculate_performance(y_true, y_pred, y_true_pops, use_best_model, bootstraps)

    def assign_unknown(self, use_best_model=True, save=True):
        """
        Assigns unknown samples to populations using the trained neural network.

        Parameters
        ----------
        use_best_model : bool, optional
            Whether to only assign samples to populations using the best model 
            (lowest validation loss during training). If set to False, then will also use all
            models generated from all training repeats and cross-validation splits to
            identify the most commonly assigned population and the frequency of assignment
            to this population. The default is True.
        save : bool, optional
            Whether to save the results to a csv file. The default is True.
        
        Returns
        -------
        pandas.DataFrame
            DataFrame containing the unknown samples and their assigned populations.
        """
        
        unknown_data = self.data.unknowns

        X_unknown = unknown_data["alleles"]
        X_unknown = _data_converter(X_unknown, None)

        if use_best_model:
            preds = self.best_model(X_unknown).argmax(axis=1)
            preds = self.label_enc.inverse_transform(preds)
            unknown_data.loc[:, "assigned_pop"] = preds

        if "bootstrap" in self.train_history.columns:
            bootstraps = self.train_history["bootstrap"].unique()
        else:
            bootstraps = None

        if not use_best_model and bootstraps is None:
            self.__pred_array = self.__assign_on_multiple_models(
                X_unknown, self.__cv_output_folder)
            
            unknown_data = self.__get_most_common_preds(unknown_data)

        elif not use_best_model:
            reps = self.train_history["rep"].unique()
            splits = self.train_history["split"].unique()
            array_width_total = len(bootstraps) * splits.max() * reps.max()
            self.__pred_array = np.zeros(shape=(len(X_unknown), array_width_total))

            for rep in reps:
                for boot in bootstraps:
                    bootstrap_folder = os.path.join(self.output_folder, f"rep{rep}_boot{boot}")
                    array_width_bootstrap = splits.max() * reps.max()
                    array_end_position = boot * array_width_bootstrap
                    array_start_position = array_end_position - array_width_bootstrap
                    new_array = self.__assign_on_multiple_models(X_unknown, bootstrap_folder)
                    self.__pred_array[:, array_start_position:array_end_position] = new_array

            unknown_data = self.__get_most_common_preds(unknown_data)

        self.__classification = unknown_data

        if save:
            unknown_data.to_csv(os.path.join(self.output_folder,
                                "classifier_assignment_results.csv"),
                                index=False)
        
        return unknown_data

    def update_unknown_samples(self, new_genetic_data, new_sample_data):
        """
        Updates the unknown samples in the classifier object.

        Parameters
        ----------
        new_genetic_data : str
            Path to the new genetic data file.
        new_sample_data : str
            Path to the new sample data file.
        
        Returns
        -------
        None.
        """
        self.__data.update_unknowns(new_genetic_data, new_sample_data)

    # Reporting functions below
    def get_classification_summary(self, save=True):
        """
        Get a summary of the classification performance metrics from running
        the test() function, including accuracy, precision, recall, and f1 
        score. Metrics are either based on the best classifier model
        (use_best_model set to True), or are averaged across the ensemble of 
        models if tested across all bootstraps, repetitions, and cross 
        validation splits (use_best_model set to False).

        Parameters
        ----------
        save : bool, optional
            Whether to save the results to a csv file. The default is True.
        
        Returns
        -------
        pandas.DataFrame
            DataFrame containing the classification summary.
        """

        summary = {
            "metric": ["accuracy", "precision", "recall", "f1"],
            "value": [self.accuracy, self.precision, self.recall, self.f1]
        }
        summary = pd.DataFrame(summary)

        if save:
            summary.to_csv(os.path.join(self.output_folder,
                          "classifier_classification_summary.csv"),
                           index=False)

        return summary
    
    def get_confusion_matrix(self):
        """
        Get the confusion matrix for the classification results.

        Returns
        -------
        numpy.ndarray
            Confusion matrix based on the results of running the test() function.
        """           
        return self.confusion_matrix

    def rank_site_importance(self, save=True):
        """
        Rank sites (SNPs) by importance in model performance.

        Parameters
        ----------
        save : bool, optional
            Whether to save the results to a csv file. The default is True.
        
        Returns
        -------
        pandas.DataFrame
            DataFrame containing the ranked sites.
        """
        if self.best_model is None:
            raise ValueError("Model has not been trained yet. " + 
            "Please run the train() method first.")

        X = self.data.knowns["alleles"].to_numpy()
        X = np.stack(X)
        Y = self.data.knowns["pop"]
        enc = OneHotEncoder(handle_unknown="ignore")
        Y_enc = enc.fit_transform(Y.values.reshape(-1, 1)).toarray()
        snp_names = np.arange(1, X.shape[1] + 1)
        errors = []

        for i in range(X.shape[1]):
            X_temp = X.copy()
            X_temp[:, i] = np.random.choice(X_temp[:, i], X_temp.shape[0])
            X_temp = torch.from_numpy(X_temp).float()
            preds = self.best_model(X_temp).argmax(axis=1)
            num_mismatches = [i for i, j in zip(preds, Y_enc.argmax(axis=1)) if i != j]
            errors.append(np.round(len(num_mismatches) / len(Y), 3))

        max_error = np.max(errors)

        if max_error == 0:
            importance = [1 for e in errors]
        else:
            importance = [1 - (1 - np.round(e / max_error, 3)) for e in errors]

        importance_data = {"snp": snp_names, "error": errors,
                           "importance": importance}
        ranking = pd.DataFrame(importance_data).sort_values("importance",
                                                            ascending=False)
        ranking.reset_index(drop=True, inplace=True)

        if save:
            ranking.to_csv(os.path.join(self.output_folder,
                          "rank_site_importance.csv"),
                           index=False)

        return ranking

    # Plotting functions below
    def plot_training_curve(self, save=True, facet_by_split_rep=False):
        """
        Plots the training curve.
        
        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        facet_by_split_rep : bool, optional
            Whether to facet the plot by split and rep. If False and more than
            1 split and rep have been used during training, then the training
            plot will contain variability corresponding to the multiple runs.
            The default is False.
            
        Returns
        -------
        None
        """

        _plot_training_curve(self.train_history, self.__nn_type,
            self.output_folder, save, facet_by_split_rep)

    def plot_confusion_matrix(self, save=True):
        """
        Plots the confusion matrix based on the results from running the test() 
        function.
        
        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        
        Returns
        -------
        None
        """
        _plot_confusion_matrix(self.test_results, self.confusion_matrix,
            self.nn_type, self.output_folder, save)

    def plot_assignment(self, save=True, col_scheme="Spectral"):
        """
        Plots the results from running the assign_unknown() function. If the 
        assign_unknown() function is run with use_best_model set to False, then plots 
        the proportion of times each sample from the unknown data was assigned to each 
        population across all bootstraps, repetitions, and cross validation splits.
        If the assign_unknown() function is run with use_best_model set to True, 
        only plots the assignment based on the results from running the data through 
        the best classifier model (all assignment frequencies will be 1).

        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        col_scheme : str, optional
            The colour scheme to use for the plot. The default is "Spectral".

        Returns
        -------
        None
        """
        if self.classification is None:
            raise ValueError("No classification results to plot.")

        if len(np.unique(self.classification.index)) == len(self.classification):
            e_preds = self.classification.copy()
            use_best_model = True

        else:
            pred_df = pd.DataFrame(self.__pred_array)
            for col in pred_df.columns:
                pred_df[col] = self.label_enc.inverse_transform(pred_df[col].astype(int))

            classifications = self.classification.copy()
            classifications.reset_index(inplace=True)
            classifications = classifications[["id"]]
            classifications = pd.concat([classifications, pred_df], axis=1)

            e_preds = pd.melt(classifications, id_vars=["id"], 
                    value_vars=pred_df.columns, 
                    value_name="assigned_pop")
            e_preds.rename(columns={"id": "sampleID"}, inplace=True)
            use_best_model = False

        _plot_assignment(e_preds, col_scheme, self.output_folder, self.__nn_type, save, use_best_model)

    def plot_structure(self, save=True, col_scheme="Spectral"):
        """
        Plots the proportion of times individuals from the
        test data were assigned to the correct population. 
        Used for determining the accuracy of the classifier.

        Parameters
        ----------
        save : bool, optional
            Whether to save the plot to a png file. The default is True.
        col_scheme : str, optional
            The colour scheme to use for the plot. The default is "Spectral".
        
        Returns
        -------
        None
        """
        preds = pd.DataFrame(self.confusion_matrix,
                            columns=self.label_enc.classes_,
                            index=self.label_enc.classes_)
        folder = self.output_folder

        _plot_structure(preds, col_scheme, self.__nn_type, folder, save)

    def save(self, save_path=None, filename="classifier.pkl"):
        """
        Saves the current instance of the class to a pickle file.

        Parameters
        ----------
        save_path : str, optional
            The path to save the file to. The default is None.
        filename : str, optional
            The name of the file to save. The default is "classifier.pkl".

        Returns
        -------
        None
        """
        _save(self, save_path, filename)

    @staticmethod
    def load(load_path=None):
        """
        Loads a saved instance of the class from a pickle file.

        Parameters
        ----------
        load_path : str, optional
            The path to load the file from. The default is None.
        
        Returns
        -------
        None
        """
        return _load(load_path)

    def _validate_init_inputs(self, data, random_state, output_folder):

        if not isinstance(data, GeneticData):
            raise TypeError("data must be an instance of GeneticData")

        if not isinstance(random_state, int):
            raise TypeError("random_state must be an integer")

        if output_folder is not None:
            if not isinstance(output_folder, str):
                raise TypeError("output_folder must be a string")

    def _validate_train_inputs(self, epochs, valid_size, cv_splits, nreps,
                               learning_rate, batch_size, dropout_prop):

        if not isinstance(epochs, int):
            raise TypeError("epochs must be an integer")
        
        if not isinstance(valid_size, float):
            raise TypeError("valid_size must be a float")

        if valid_size > 1 or valid_size < 0:
            raise ValueError("valid_size must be between 0 and 1")
        
        if not isinstance(cv_splits, int):
            raise TypeError("cv_splits must be an integer")

        if not isinstance(nreps, int):
            raise TypeError("nreps must be an integer")

        if not isinstance(learning_rate, float):
            raise TypeError("learning_rate must be a float")

        if learning_rate > 1 or learning_rate < 0:
            raise ValueError("learning_rate must be between 0 and 1")

        if not isinstance(batch_size, int):
            raise TypeError("batch_size must be an integer")

        if not isinstance(dropout_prop, float) and not isinstance(dropout_prop, int):
            raise TypeError("dropout_prop must be a float")

        if dropout_prop > 1 or dropout_prop < 0:
            raise ValueError("dropout_prop must be between 0 and 1")

    # Hidden functions below   
    def __train_on_inputs(self, inputs, cv_splits, epochs, learning_rate, batch_size, 
                          dropout_prop, result_folder, overwrite_results=True):

        self.__prepare_result_folder(result_folder, overwrite_results)

        loss_dict = {"split": [], "epoch": [], "train": [], "valid": []}

        for i, input in enumerate(inputs):

            lowest_val_loss_rep = 9999
            split = i % cv_splits + 1

            X_train, y_train, X_valid, y_valid = _split_input_classifier(self, input)
            train_loader, valid_loader = _generate_data_loaders(X_train, y_train,
                                                                X_valid, y_valid)

            net = ClassifierNet(input_size=X_train.shape[1], hidden_size=16, #TODO: make hidden size a parameter
                                output_size=len(y_train.unique()),
                                batch_size=batch_size, dropout_prop=dropout_prop)
            optimizer = torch.optim.Adam(net.parameters(), lr=learning_rate)
            loss_func = nn.CrossEntropyLoss()

            for epoch in range(epochs):

                train_loss = 0
                valid_loss = 0

                for _, (data, target) in enumerate(train_loader):
                    optimizer.zero_grad()
                    output = net(data)
                    loss = loss_func(output.squeeze(), target.squeeze().long())
                    loss.backward()
                    optimizer.step()
                    train_loss += loss.data.item()
            
                # Calculate average train loss
                avg_train_loss = train_loss / len(train_loader)

                for _, (data, target) in enumerate(valid_loader):
                    output = net(data)
                    loss = loss_func(output.squeeze(), target.squeeze().long())
                    valid_loss += loss.data.item()

                    if valid_loss < lowest_val_loss_rep:
                        lowest_val_loss_rep = valid_loss
                        torch.save(net, os.path.join(result_folder, f"best_model_split{split}.pt"))

                    if valid_loss < self.__lowest_val_loss_total:
                        self.__lowest_val_loss_total = valid_loss
                        torch.save(net, os.path.join(self.output_folder, "best_model.pt"))

                # Calculate average validation loss
                avg_valid_loss = valid_loss / len(valid_loader)

                loss_dict["split"].append(split)
                loss_dict["epoch"].append(epoch)
                loss_dict["train"].append(avg_train_loss)
                loss_dict["valid"].append(avg_valid_loss)

        return pd.DataFrame(loss_dict)
    
    def __prepare_result_folder(self, result_folder, overwrite_results=True):

        # Make result folder if it doesn't exist
        if not os.path.exists(result_folder):
            os.mkdir(result_folder)
        elif overwrite_results:
            shutil.rmtree(result_folder)
            os.mkdir(result_folder)

    def __test_on_multiple_models(self, reps, bootstraps, splits, X_test, y_true_pops):

        result_df = pd.DataFrame()
        for rep in reps:
            for boot in bootstraps:
                for split in splits:
                    folder = os.path.join(self.output_folder, f"rep{rep}_boot{boot}")
                    model = torch.load(os.path.join(folder, f"best_model_split{split}.pt"))
                    y_pred = model(X_test).argmax(axis=1)
                    y_pred_pops = self.label_enc.inverse_transform(y_pred)
                    cv_test_results_temp = pd.DataFrame(
                        {"rep": rep, "bootstrap": boot, "split": split, 
                        "true_pop": y_true_pops, "pred_pop": y_pred_pops})
                    result_df = pd.concat([result_df, cv_test_results_temp])

        return result_df
                                    

    def __calculate_performance(self, y_true, y_pred, y_true_pops, use_best_model, bootstraps):

        # Calculate ensemble performance metrics if not best model only
        if not use_best_model and bootstraps is None:
            self.__test_on_cv_splits = True

        elif not use_best_model:
            self.__test_on_bootstraps = True

        results = self.__organize_performance_metrics(self.test_results, y_true_pops, y_true, y_pred)
        self.__confusion_matrix, self.__accuracy, self.__precision, self.__recall, self.__f1 = results

    def __organize_performance_metrics(self, result_df, y_true_pops, y_true, y_pred):
        cf = np.round(confusion_matrix(
            result_df["true_pop"], result_df["pred_pop"], 
            labels=np.unique(y_true_pops).tolist(), normalize="true"), 3)
        accuracy = np.round(accuracy_score(y_true, y_pred), 3)
        precision = np.round(precision_score(y_true, y_pred, average="weighted"), 3)
        recall = np.round(recall_score(y_true, y_pred, average="weighted"), 3)
        f1 = np.round(f1_score(y_true, y_pred, average="weighted"), 3)

        return cf, accuracy, precision, recall, f1

    def __assign_on_multiple_models(self, X_unknown, folder):
        reps = self.train_history["rep"].unique()
        splits = self.train_history["split"].unique()

        # Create empty array to fill
        array_width_total = splits.max() * reps.max()
        array = np.zeros(shape=(len(X_unknown), array_width_total))
        pos = 0

        for rep in reps:
            for split in splits:
                model = torch.load(os.path.join(folder, f"best_model_split{split}.pt"))
                preds = model(X_unknown).argmax(axis=1)
                array[:, pos] = preds
                pos += 1

        return array

    def __get_most_common_preds(self, unknown_data):
        """
        Want to retrieve the most common prediction across all reps / splits
        for each unknown sample - give estimate of confidence based on how
        many times a sample is assigned to a population
        """
        most_common = np.array([Counter(sorted(row, reverse=True)).\
                                most_common(1)[0][0] for row in self.__pred_array])
        most_common_count = np.count_nonzero(self.__pred_array == most_common[:, None], axis=1)
        frequency = np.round(most_common_count / self.__pred_array.shape[1], 3)
        most_common = self.label_enc.inverse_transform(most_common.astype(int))
        unknown_data.loc[:, "most_assigned_pop_across_models"] = most_common    
        unknown_data.loc[:, "frequency_of_assignment_across_models"] = frequency

        return unknown_data





import pandas as pd
import numpy as np
import multiprocessing as mp
import argparse
import os

from dataloader import GeneticData
from classifier import PopClassifier
# from popfinder.regressor import PopRegressor

def _train_on_bootstraps(clf_object, train_args):
    
    # Train on new training set
    clf_object.train(epochs=train_args["epochs"],
                            valid_size=train_args["valid_size"],
                            cv_splits=train_args["cv_splits"],
                            learning_rate=train_args["learning_rate"],
                            batch_size=train_args["batch_size"],
                            dropout_prop=train_args["dropout_prop"])
    # Save trained model to output folder
    # clf_object.save()

    # Return losses
    clf_object.train_history.to_csv(os.path.join(clf_object.output_folder, "loss.csv"), index=False)
    return clf_object.train_history

def create_classifier_objects(rep_start, nreps, nboots, popfinder_path):

    classifier_objects = []
    for rep in range(rep_start, nreps):
        for boot in range(nboots):

            popfinder = PopClassifier.load(os.path.join(popfinder_path, "classifier.pkl"))
            popfinder.output_folder = os.path.join(popfinder.output_folder, f"rep{rep+1}_boot{boot+1}")
            os.makedirs(popfinder.output_folder, exist_ok=True)

            # Use bootstrap to randomly select sites from training/test/unknown data
            num_sites = popfinder.data.train["alleles"].values[0].shape[0]

            site_indices = np.random.choice(range(num_sites), size=num_sites,
                                            replace=True)

            popfinder.__boot_data = GeneticData()
            popfinder.__boot_data.train = popfinder.data.train.copy()
            popfinder.__boot_data.test = popfinder.data.test.copy()
            popfinder.__boot_data.knowns = pd.concat([popfinder.data.train, popfinder.data.test])
            popfinder.__boot_data.unknowns = popfinder.data.unknowns.copy()

            # Slice datasets by site_indices
            popfinder.__boot_data.train["alleles"] = [a[site_indices] for a in popfinder.data.train["alleles"].values]
            popfinder.__boot_data.test["alleles"] = [a[site_indices] for a in popfinder.data.test["alleles"].values]
            popfinder.__boot_data.unknowns["alleles"] = [a[site_indices] for a in popfinder.data.unknowns["alleles"].values]

            classifier_objects.append(popfinder)

    return classifier_objects

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", help="Path to PopClassifier object")
    parser.add_argument("-n", help="Number of bootstraps", type=int)
    parser.add_argument("--r_start", help="Starting repetition number", type=int)
    parser.add_argument("-r", help="Number of repetitions", type=int)
    parser.add_argument("-e", help="Number of epochs", type=int)
    parser.add_argument("-v", help="Validation size", type=float)
    parser.add_argument("-s", help="Number of cross-validation splits", type=int)
    parser.add_argument("-l", help="Learning rate", type=float)
    parser.add_argument("-b", help="Batch size", type=int)
    parser.add_argument("-d", help="Dropout proportion", type=float)
    parser.add_argument("-j", help="Number of jobs", type=int)
    args = parser.parse_args()
    popfinder_path = args.p
    nboots = args.n
    rep_start = args.r_start
    nreps = args.r
    epochs = args.e 
    valid_size = args.v 
    cv_splits = args.s
    learning_rate = args.l 
    batch_size = args.b
    dropout_prop = args.d
    num_jobs = args.j

    # Generate inputs
    classifier_objects = create_classifier_objects(rep_start, nreps, nboots, popfinder_path)
    # Create dictionary of train args
    train_args = {"epochs": epochs, "valid_size": valid_size, "cv_splits": cv_splits,
                  "learning_rate": learning_rate, "batch_size": batch_size,
                  "dropout_prop": dropout_prop}

    if num_jobs == -1:
        num_jobs = mp.cpu_count()
    pool = mp.Pool(processes=num_jobs)
    results = pool.starmap(_train_on_bootstraps, [(c, train_args) for c in classifier_objects])
    pool.close()
    pool.join()

    for rep in range((nreps - rep_start)):
        for boot in range(nboots):
            ind = rep * nboots + boot
            results[ind]["rep"] = rep_start + rep + 1
            results[ind]["bootstrap"] = boot + 1
    
    final_results = pd.concat(results, ignore_index=True)
    final_results.to_csv(os.path.join(popfinder_path, "train_history.csv"), index=False)    